import fnmatch
from errbot import BotPlugin, cmdfilter
from errbot.backends.base import RoomOccupant

BLOCK_COMMAND = (None, None, None)


def get_acl_usr(msg):
    """Return the ACL attribute of the sender of the given message"""
    if hasattr(msg.frm, 'aclattr'):  # if the identity requires a special field to be used for acl
        return msg.frm.aclattr
    return msg.frm.person  # default


def glob(text, patterns):
    """
    Match text against the list of patterns according to unix glob rules.
    Return True if a match is found, False otherwise.
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    if not isinstance(text, str):
        text = str(text)
    return any(fnmatch.fnmatchcase(text, str(pattern)) for pattern in patterns)


def ciglob(text, patterns):
    """
    Case-insensitive version of glob.

    Match text against the list of patterns according to unix glob rules.
    Return True if a match is found, False otherwise.
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    return glob(text.lower(), [p.lower() for p in patterns])


class ACLS(BotPlugin):
    """
    This plugin implements access controls for commands, allowing them to be
    restricted via various rules.
    """

    def access_denied(self, msg, reason, dry_run):
        if not dry_run and not self.bot_config.HIDE_RESTRICTED_ACCESS:
            self._bot.send_simple_reply(msg, reason)
        return BLOCK_COMMAND

    @cmdfilter
    def acls(self, msg, cmd, args, dry_run):
        """
        Check command against ACL rules as defined in the bot configuration.

        :param msg: The original chat message.
        :param cmd: The command name itself.
        :param args: Arguments passed to the command.
        :param dry_run: True when this is a dry-run.
        """
        self.log.debug('Check %s for ACLs.', cmd)
        f = self._bot.all_commands[cmd]
        cmd_str = f'{f.__self__.name}:{cmd}'

        usr = get_acl_usr(msg)
        acl = self.bot_config.ACCESS_CONTROLS_DEFAULT.copy()
        for pattern, acls in self.bot_config.ACCESS_CONTROLS.items():
            if ':' not in pattern:
                pattern = f'*:{pattern}'
            if ciglob(cmd_str, (pattern,)):
                acl.update(acls)
                break

        self.log.info('Matching ACL %s against username %s for command %s.', acl, usr, cmd_str)

        if 'allowusers' in acl and not glob(usr, acl['allowusers']):
            return self.access_denied(msg, "You're not allowed to access this command from this user", dry_run)
        if 'denyusers' in acl and glob(usr, acl['denyusers']):
            return self.access_denied(msg, "You're not allowed to access this command from this user", dry_run)
        if msg.is_group:
            if not isinstance(msg.frm, RoomOccupant):
                raise Exception(f'msg.frm is not a RoomOccupant. Class of frm: {msg.frm.__class__}')
            room = str(msg.frm.room)
            if 'allowmuc' in acl and acl['allowmuc'] is False:
                return self.access_denied(msg, "You're not allowed to access this command from a chatroom", dry_run)

            if 'allowrooms' in acl and not glob(room, acl['allowrooms']):
                return self.access_denied(msg, "You're not allowed to access this command from this room", dry_run)

            if 'denyrooms' in acl and glob(room, acl['denyrooms']):
                return self.access_denied(msg, "You're not allowed to access this command from this room", dry_run)
        elif 'allowprivate' in acl and acl['allowprivate'] is False:
            return self.access_denied(
                msg,
                "You're not allowed to access this command via private message to me",
                dry_run
            )

        self.log.debug('Check if %s is admin only command.', cmd)
        if f._err_command_admin_only:
            if not glob(get_acl_usr(msg), self.bot_config.BOT_ADMINS):
                return self.access_denied(msg, 'This command requires bot-admin privileges', dry_run)
            # For security reasons, admin-only commands are direct-message only UNLESS
            # specifically overridden by setting allowmuc to True for such commands.
            if msg.is_group and not acl.get('allowmuc', False):
                return self.access_denied(msg, 'This command may only be issued through a direct message', dry_run)

        return msg, cmd, args



import os
import pytest
import tempfile
from configparser import ConfigParser
from pathlib import Path

import errbot.repo_manager
from errbot import plugin_manager
from errbot.plugin_info import PluginInfo
from errbot.plugin_manager import IncompatiblePluginException
from errbot.utils import find_roots, collect_roots

CORE_PLUGINS = plugin_manager.CORE_PLUGINS


def touch(name):
    with open(name, 'a'):
        os.utime(name, None)


assets = Path(__file__).parent / 'assets'


def test_check_dependencies():
    response, deps = errbot.repo_manager.check_dependencies(assets / 'requirements_never_there.txt')
    assert 'You need these dependencies for' in response
    assert 'impossible_requirement' in response
    assert ['impossible_requirement'] == deps


def test_check_dependencies_no_requirements_file():
    response, deps = errbot.repo_manager.check_dependencies(assets / 'requirements_non_existent.txt')
    assert response is None


def test_check_dependencies_requirements_file_all_installed():
    response, deps = errbot.repo_manager.check_dependencies(assets / 'requirements_already_there.txt')
    assert response is None


def test_find_plugin_roots():
    root = tempfile.mkdtemp()
    a = os.path.join(root, 'a')
    b = os.path.join(a, 'b')
    c = os.path.join(root, 'c')
    os.mkdir(a)
    os.mkdir(b)
    os.mkdir(c)
    touch(os.path.join(a, 'toto.plug'))
    touch(os.path.join(b, 'titi.plug'))
    touch(os.path.join(root, 'tutu.plug'))
    roots = find_roots(root)
    assert root in roots
    assert a in roots
    assert b in roots
    assert c not in roots


def test_collect_roots():
    toto = tempfile.mkdtemp()
    touch(os.path.join(toto, 'toto.plug'))
    touch(os.path.join(toto, 'titi.plug'))
    titi = tempfile.mkdtemp()
    touch(os.path.join(titi, 'tata.plug'))
    tutu = tempfile.mkdtemp()
    subtutu = os.path.join(tutu, 'subtutu')
    os.mkdir(subtutu)
    touch(os.path.join(subtutu, 'tutu.plug'))

    assert collect_roots((CORE_PLUGINS, None)) == [CORE_PLUGINS]
    assert collect_roots((CORE_PLUGINS, toto)) == [CORE_PLUGINS, toto]
    assert collect_roots((CORE_PLUGINS, [toto, titi])) == [CORE_PLUGINS, toto, titi]
    assert collect_roots((CORE_PLUGINS, toto, titi, 'nothing')) == [CORE_PLUGINS, toto, titi]
    assert collect_roots((toto, tutu)) == [toto, subtutu]


def test_ignore_dotted_directories():
    root = tempfile.mkdtemp()
    a = os.path.join(root, '.invisible')
    os.mkdir(a)
    touch(os.path.join(a, 'toto.plug'))
    assert collect_roots((CORE_PLUGINS, root)) == [CORE_PLUGINS]


def dummy_config_parser() -> ConfigParser:
    cp = ConfigParser()
    cp.add_section('Core')
    cp.set('Core', 'Name', 'dummy')
    cp.set('Core', 'Module', 'dummy')
    cp.add_section('Errbot')
    return cp


def test_errbot_version_check():
    real_version = plugin_manager.VERSION

    too_high_min_1 = dummy_config_parser()
    too_high_min_1.set('Errbot', 'Min', '1.6.0')

    too_high_min_2 = dummy_config_parser()
    too_high_min_2.set('Errbot', 'Min', '1.6.0')
    too_high_min_2.set('Errbot', 'Max', '2.0.0')

    too_low_max_1 = dummy_config_parser()
    too_low_max_1.set('Errbot', 'Max', '1.0.1-beta')

    too_low_max_2 = dummy_config_parser()
    too_low_max_2.set('Errbot', 'Min', '0.9.0-rc2')
    too_low_max_2.set('Errbot', 'Max', '1.0.1-beta')

    ok1 = dummy_config_parser()  # empty section

    ok2 = dummy_config_parser()
    ok2.set('Errbot', 'Min', '1.4.0')

    ok3 = dummy_config_parser()
    ok3.set('Errbot', 'Max', '1.5.2')

    ok4 = dummy_config_parser()
    ok4.set('Errbot', 'Min', '1.2.1')
    ok4.set('Errbot', 'Max', '1.6.1-rc1')

    try:
        plugin_manager.VERSION = '1.5.2'
        for config in (too_high_min_1,
                       too_high_min_2,
                       too_low_max_1,
                       too_low_max_2):
            pi = PluginInfo.parse(config)
            with pytest.raises(IncompatiblePluginException):
                plugin_manager.check_errbot_version(pi)

        for config in (ok1, ok2, ok3, ok4):
            pi = PluginInfo.parse(config)
            plugin_manager.check_errbot_version(pi)
    finally:
        plugin_manager.VERSION = real_version



#!/usr/bin/env python3
import json

from jinja2 import Template

template = Template(open('plugins.md').read())

blacklisted = [repo.strip() for repo in open('blacklisted.txt', 'r').readlines()]

PREFIX_LEN = len('https://github.com/')

with open('repos.json', 'r') as p:
    repos = json.load(p)

    # Removes the weird forks of errbot itself and
    # blacklisted repos
    filtered_plugins = []
    for repo, plugins in repos.items():
        for name, plugin in plugins.items():
            if plugin['path'].startswith('errbot/builtins'):
                continue
            if plugin['repo'][PREFIX_LEN:] in blacklisted:
                continue
            filtered_plugins.append(plugin)

    sorted_plugins = sorted(filtered_plugins, key=lambda plugin: -plugin['score'])

    with open('Home.md', 'w') as out:
        out.write(template.render(plugins=sorted_plugins))



##########################################################################
#                                                                        #
#  This is the config-template for Err. This file should be copied and   #
#  renamed to config.py, then modified as you see fit to run Errbot      #
#  the way you like it.                                                  #
#                                                                        #
#  As this is a regular Python file, note that you can do variable       #
#  assignments and the likes as usual. This can be useful for example if #
#  you use the same values in multiple places.                           #
#                                                                        #
#  Note: Various config options require a tuple to be specified, even    #
#  when you are configuring only a single value. An example of this is   #
#  the BOT_ADMINS option. Make sure you use a valid tuple here, even if  #
#  you are only configuring a single item, else you will get errors.     #
#  (So don't forget the trailing ',' in these cases)                     #
#                                                                        #
##########################################################################

import logging

##########################################################################
# Core Errbot configuration                                              #
##########################################################################

# BACKEND selection.
# This configures the type of chat server you wish to use Errbot with.
#
# The current choices:

# Debug backends to test your plugins manually:
# 'Text'     - on the text console
# 'Graphic'  - in a GUI window

# Commercial backends:
# 'Campfire' - see https://campfirenow.com/ (follow instructions from https://github.com/errbotio/err-backend-campfire)
# 'Hipchat'  - see https://www.hipchat.com/
# 'Slack'    - see https://slack.com/
# 'Gitter'   - see https://gitter.im/ (follow instructions from https://github.com/errbotio/err-backend-gitter)

# Open protocols:
# 'TOX'      - see https://tox.im/ (follow instructions from https://github.com/errbotio/err-backend-tox)
# 'IRC'      - for classic IRC or bridged services like https://gitter.im
# 'XMPP'     - the Extensible Messaging and Presence Protocol (https://xmpp.org/)
# 'Telegram' - cloud-based mobile and desktop messaging app with a focus
#              on security and speed. (https://telegram.org/)

# BACKEND = 'XMPP'  # defaults to XMPP

# STORAGE selection.
# This configures the type of persistence you wish to use Errbot with.
#
# The current choices:

# Debug:
# 'Memory'        - local memory storage to test your bot in memory:

# Filesystem:
# 'Shelf'         - python shelf (default)

# STORAGE = 'Shelf'  # defaults to filestorage (python shelf).

# BOT_EXTRA_STORAGE_PLUGINS_DIR = None  # extra search path to find custom storage plugins

# The location where all of Err's data should be stored. Make sure to set
# this to a directory that is writable by the user running the bot.
BOT_DATA_DIR = '/var/lib/err'

### Repos and plugins config.

# Set this to change from where errbot gets its installable plugin list.
# By default it gets the index from errbot.io which is a file generated by tools/plugin-gen.py.
# BOT_PLUGIN_INDEXES = 'https://errbot.io/repos.json'
#
# You can also specify a local file:
# BOT_PLUGIN_INDEXES = 'tools/repos.json'
#
# Or a list. note: if some plugins exists in 2 lists, only the first hit will be taken into account.
# BOT_PLUGIN_INDEXES = ('/data/repos.json', 'https://my.private.tld/errbot/myrepos.json')

# Set this to a directory on your system where you want to load extra
# plugins from, which is useful mostly if you want to develop a plugin
# locally before publishing it. Note that you can specify only a single
# directory, however you are free to create subdirectories with multiple
# plugins inside this directory.
BOT_EXTRA_PLUGIN_DIR = None

# If you use an external backend as a plugin,
# this is where you tell Errbot where to find it.
# BOT_EXTRA_BACKEND_DIR = '/opt/errbackends'

# If you want only a subset of the core plugins that are bundled with errbot, you can specify them here.
# CORE_PLUGINS = None # This is default, all core plugins.
# For example CORE_PLUGINS = ('ACLs', 'Backup', 'Help') you get those names from the .plug files Name entry.
# For absolutely no plug: CORE_PLUGINS = ()

# Defines an order in which the plugins are getting their callbacks. Useful if you want to have plugins do
# pre- or post-processing on messages.
# The 'None' tuple entry represents all the plugins that aren't to be explicitly ordered. For example, if
# you want 'A' to run first, then everything else but 'B', then 'B', you would use ('A', None, 'B').
PLUGINS_CALLBACK_ORDER = (None, )

# Should plugin dependencies be installed automatically? If this is true
# then Errbot will use pip to install any missing dependencies automatically.
#
# If you have installed Errbot in a virtualenv, this will run the equivalent
# of `pip install -r requirements.txt`.
# If no virtualenv is detected, the equivalent of `pip install --user -r
# requirements.txt` is used to ensure the package(s) is/are only installed for
# the user running Err.
#AUTOINSTALL_DEPS = True

# To use your own custom log formatter, uncomment and set BOT_LOG_FORMATTER
# to your formatter instance (inherits from logging.Formatter)
#   For information on how to create a logging formatter and what it can do, see
#   https://docs.python.org/3/library/logging.html#formatter-objects
# BOT_LOG_FORMATTER =

# The location of the log file. If you set this to None, then logging will
# happen to console only.
BOT_LOG_FILE = BOT_DATA_DIR + '/err.log'

# The verbosity level of logging that is done to the above logfile, and to
# the console. This takes the standard Python logging levels, DEBUG, INFO,
# WARN, ERROR. For more info, see http://docs.python.org/library/logging.html
#
# If you encounter any issues with Err, please set your log level to
# logging.DEBUG and attach a log with your bug report to aid the developers
# in debugging the issue.
BOT_LOG_LEVEL = logging.INFO

# Enable logging to sentry (find out more about sentry at www.getsentry.com).
# You can also separate Flask exceptions by enabling it. This will give more information in sentry
# This is optional and disabled by default.
BOT_LOG_SENTRY = False
BOT_LOG_SENTRY_FLASK = False
SENTRY_DSN = ''
SENTRY_LOGLEVEL = BOT_LOG_LEVEL
SENTRY_EVENTLEVEL = BOT_LOG_LEVEL

# Set an optional Sentry transport other than the default Threaded.
# For more info, see https://docs.sentry.io/error-reporting/configuration/?platform=python#transport-options
# SENTRY_TRANSPORT = ('RequestsHTTPTransport', 'raven.transport.requests')

# Execute commands in asynchronous mode. In this mode, Errbot will spawn 10
# separate threads to handle commands, instead of blocking on each
# single command.
# BOT_ASYNC = True

# Size of the thread pool for the asynchronous mode.
# BOT_ASYNC_POOLSIZE = 10

##########################################################################
# Account and chatroom (MUC) configuration                               #
##########################################################################

# The identity, or credentials, used to connect to a server
BOT_IDENTITY = {
    # XMPP (Jabber) mode
    'username': 'err@localhost',  # The JID of the user you have created for the bot
    'password': 'changeme',       # The corresponding password for this user
    # 'server': ('host.domain.tld',5222), # server override

    ## HipChat mode (Comment the above if using this mode)
    # 'username': '12345_123456@chat.hipchat.com',
    # 'password': 'changeme',
    ## Group admins can create/view tokens on the settings page after logging
    ## in on HipChat's website
    # 'token': 'ed4b74d62833267d98aa99f312ff04',
    ## If you're using HipChat server (self-hosted HipChat) then you should set
    ## the endpoint below. If you don't use HipChat server but use the hosted version
    ## of HipChat then you may leave this commented out.
    # 'endpoint': 'https://api.hipchat.com'

    ## Slack mode (comment the others above if using this mode)
    # 'token': 'xoxb-4426949411-aEM7...',
    ## you can also include the proxy for the SlackClient connection
    # 'proxies': {'http': 'some-http-proxy', 'https': 'some-https-proxy'}

    ## Telegram mode (comment the others above if using this mode)
    # 'token': '103419016:AAbcd1234...',

    ## IRC mode (Comment the others above if using this mode)
    # 'nickname': 'err-chatbot',
    # 'username': 'err-chatbot',    # optional, defaults to nickname if omitted
    # 'password': None,             # optional
    # 'server': 'irc.freenode.net',
    # 'port': 6667,                  # optional
    # 'ssl': False,                  # optional
    # 'ipv6': False,                 # optional
    # 'nickserv_password': None,     # optional
    ## Optional: Specify an IP address or hostname (vhost), and a
    ## port, to use when making the connection. Leave port at 0
    ## if you have no source port preference.
    ##    example: 'bind_address': ('my-errbot.io', 0)
    # 'bind_address': ('localhost', 0),
}

# Set the admins of your bot. Only these users will have access
# to the admin-only commands.
#
# Unix-style glob patterns are supported, so 'gbin@localhost'
# would be considered an admin if setting '*@localhost'.
BOT_ADMINS = ('gbin@localhost',)

# Set of admins that wish to receive administrative bot notifications.
#BOT_ADMINS_NOTIFICATIONS = ()

# Chatrooms your bot should join on startup. For the IRC backend you
# should include the # sign here. For XMPP rooms that are password
# protected, you can specify another tuple here instead of a string,
# using the format (RoomName, Password).
# CHATROOM_PRESENCE = ('err@conference.server.tld',)

# The FullName, or nickname, your bot should use. What you set here will
# be the nickname that Errbot shows in chatrooms. Note that some XMPP
# implementations, notably HipChat, are very picky about what name you
# use. In the case of HipChat, make sure this matches exactly with the
# name you gave the user.
# CHATROOM_FN = 'Errbot'

##########################################################################
# Prefix configuration                                                   #
##########################################################################

# Command prefix, the prefix that is expected in front of commands directed
# at the bot.
#
# Note: When writing plugins,you should always use the default '!'.
# If the prefix is changed from the default, the help strings will be
# automatically adjusted for you.
#
# BOT_PREFIX = '!'
#
# Uncomment the following and set it to True if you want the prefix to be
# optional for normal chat.
# (Meaning messages sent directly to the bot as opposed to within a MUC)
#BOT_PREFIX_OPTIONAL_ON_CHAT = False

# You might wish to have your bot respond by being called with certain
# names, rather than the BOT_PREFIX above. This option allows you to
# specify alternative prefixes the bot will respond to in addition to
# the prefix above.
#BOT_ALT_PREFIXES = ('Err',)

# If you use alternative prefixes, you might want to allow users to insert
# separators like , and ; between the prefix and the command itself. This
# allows users to refer to your bot like this (Assuming 'Err' is in your
# BOT_ALT_PREFIXES):
# "Err, status" or "Err: status"
#
# Note: There's no need to add spaces to the separators here
#
#BOT_ALT_PREFIX_SEPARATORS = (':', ',', ';')

# Continuing on this theme, you might want to permit your users to be
# lazy and not require correct capitalization, so they can do 'Err',
# 'err' or even 'ERR'.
#BOT_ALT_PREFIX_CASEINSENSITIVE = True

##########################################################################
# Access controls and message diversion                                  #
##########################################################################

# Access controls, allowing commands to be restricted to specific users/rooms.
# Available filters (you can omit a filter or set it to None to disable it):
#   allowusers: Allow command from these users only
#   denyusers: Deny command from these users
#   allowrooms: Allow command only in these rooms (and direct messages)
#   denyrooms: Deny command in these rooms
#   allowprivate: Allow command from direct messages to the bot
#   allowmuc: Allow command inside rooms
# Rules listed in ACCESS_CONTROLS_DEFAULT are applied by default and merged
# with any commands found in ACCESS_CONTROLS.
#
# The options allowusers, denyusers, allowrooms and denyrooms support
# unix-style globbing similar to BOT_ADMINS.
#
# Command names also support unix-style globs and can optionally be restricted
# to a specific plugin by prefixing the command with the name of a plugin,
# separated by a colon. For example, `Health:status` will match the `!status`
# command of the `Health` plugin and `Health:*` will match all commands defined
# by the `Health` plugin.
#
# Please note that the first command match found will be used so if you have
# overlapping patterns you must used an OrderedDict instead of a regular dict:
# https://docs.python.org/3/library/collections.html#collections.OrderedDict
#
# Example:
#
#ACCESS_CONTROLS_DEFAULT = {} # Allow everyone access by default
#ACCESS_CONTROLS = {'status': {'allowrooms': ('someroom@conference.localhost',)},
#                   'about': {'denyusers': ('*@evilhost',), 'allowrooms': ('room1@conference.localhost', 'room2@conference.localhost')},
#                   'uptime': {'allowusers': BOT_ADMINS},
#                   'help': {'allowmuc': False},
#                   'help': {'allowmuc': False},
#                   'ChatRoom:*': {'allowusers': BOT_ADMINS},
#                  }

# Uncomment and set this to True to hide the restricted commands from
# the help output.
#HIDE_RESTRICTED_COMMANDS = False

# Uncomment and set this to True to ignore commands from users that have no
# access for these instead of replying with error message.
#HIDE_RESTRICTED_ACCESS = False

# A list of commands which should be responded to in private, even if
# the command was given in a MUC. For example:
# DIVERT_TO_PRIVATE = ('help', 'about', 'status')
DIVERT_TO_PRIVATE = ()

# A list of commands which should be responded to in a thread if the backend supports it.
# For example:
# DIVERT_TO_THREAD = ('help', 'about', 'status')
DIVERT_TO_THREAD = ()

# Chat relay
# Can be used to relay one to one message from specific users to the bot
# to MUCs. This can be useful with XMPP notifiers like for example the
# standard Altassian Jira which don't have native support for MUC.
# For example: CHATROOM_RELAY = {'gbin@localhost' : (_TEST_ROOM,)}
CHATROOM_RELAY = {}

# Reverse chat relay
# This feature forwards whatever is said to a specific user.
# It can be useful if you client like gtalk doesn't support MUC correctly
# For example: REVERSE_CHATROOM_RELAY = {_TEST_ROOM : ('gbin@localhost',)}
REVERSE_CHATROOM_RELAY = {}

##########################################################################
# Miscellaneous configuration options                                    #
##########################################################################

# Define the maximum length a single message may be. If a plugin tries to
# send a message longer than this length, it will be broken up into multiple
# shorter messages that do fit.
#MESSAGE_SIZE_LIMIT = 10000

# XMPP TLS certificate verification. In order to validate offered certificates,
# you must supply a path to a file containing certificate authorities. By
# default, "/etc/ssl/certs/ca-certificates.crt" is used, which on most Linux
# systems holds the default system trusted CA certificates. You might need to
# change this depending on your environment. Setting this to None disables
# certificate validation, which can be useful if you have a self-signed
# certificate for example.
#XMPP_CA_CERT_FILE = "/etc/ssl/certs/ca-certificates.crt"

# Influence the security methods used on connection with XMPP-based
# backends. You can use this to work around authentication issues with
# some buggy XMPP servers.
#
# The default is to try anything:
#XMPP_FEATURE_MECHANISMS = {}
# To use only unencrypted plain auth:
#XMPP_FEATURE_MECHANISMS = {'use_mech': 'PLAIN', 'unencrypted_plain': True, 'encrypted_plain': False}

# Modify the default keep-alive interval. By default, Errbot will send
# some whitespace to the XMPP server every 300 seconds to keep the TCP
# connection alive. On some servers, or when running Errbot from behind
# a NAT router, the default might not be fast enough and you will need
# to set it to a lower value.
#
# It has been reported that HipChat also times out without setting this
# to a lower value (60 seems to work well with HipChat)
#
# If you're having issues with your bot getting constantly disconnected,
# try to gradually lower this value until it no longer happens.
#XMPP_KEEPALIVE_INTERVAL = 300

# Modify default settings for IPv6 usage. This key affect both
# XMPP and HipChat backend.
#XMPP_USE_IPV6 = False

# XMPP supports some formatting with XEP-0071 (http://www.xmpp.org/extensions/xep-0071.html).
# It is disabled by default because XMPP clients support has been found to be spotty.
# Switch it to True to enable XHTML-IM formatting.
# XMPP_XHTML_IM = False

# Message rate limiting for the IRC backend. This will delay subsequent
# messages by this many seconds (floats are supported). Setting these
# to a value of 0 effectively disables rate limiting.
#IRC_CHANNEL_RATE = 1  # Regular channel messages
#IRC_PRIVATE_RATE = 1  # Private messages
#IRC_RECONNECT_ON_KICK = 5  # Reconnect back to a channel after a kick (in seconds)
                            # Put it at None if you don't want the chat to
                            # reconnect
#IRC_RECONNECT_ON_DISCONNECT = 5  # Reconnect back to a channel after a disconnection (in seconds)

# The pattern to build a user representation from for ACL matches.
# The default is "{nick}!{user}@{host}" which results in "zoni!zoni@ams1.groenen.me"
# for the user zoni connecting from ams1.groenen.me.
# Available substitution variables:
#   {nick}  ->  The nickname of the user
#   {user}  ->  The username of the user
#   {host}  ->  The hostname the user is connecting from
#IRC_ACL_PATTERN = "{nick}!{user}@{host}"

# Allow messages sent in a chatroom to be directed at requester.
#GROUPCHAT_NICK_PREFIXED = False

# Disable table borders, making output more compact (supported only on IRC, Slack and Telegram currently).
# COMPACT_OUTPUT = True

# Disables the logging output in Text mode and only outputs Ansi.
# TEXT_DEMO_MODE = False

# Prevent ErrBot from saying anything if the command is unrecognized.
# SUPPRESS_CMD_NOT_FOUND = False



from typing import Tuple, Union, Sequence, List, Generator, Dict

import json
import re
import logging
import os
import shutil
import subprocess
from collections import namedtuple
from datetime import timedelta, datetime
from os import path
import tarfile
from pathlib import Path
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request
from urllib.parse import urlparse

from errbot.storage import StoreMixin
from errbot.storage.base import StoragePluginBase
from .utils import git_clone
from .utils import git_pull
from .utils import ON_WINDOWS

log = logging.getLogger(__name__)


def human_name_for_git_url(url):
    # try to humanize the last part of the git url as much as we can
    s = url.split(':')[-1].split('/')[-2:]
    if s[-1].endswith('.git'):
        s[-1] = s[-1][:-4]
    return str('/'.join(s))


INSTALLED_REPOS = 'installed_repos'

REPO_INDEXES_CHECK_INTERVAL = timedelta(hours=1)

REPO_INDEX = 'repo_index'
LAST_UPDATE = 'last_update'

RepoEntry = namedtuple('RepoEntry', 'entry_name, name, python, repo, path, avatar_url, documentation')
FIND_WORDS_RE = re.compile(r"(\w[\w']*\w|\w)")


class RepoException(Exception):
    pass


def makeEntry(repo_name: str, plugin_name: str, json_value):
    return RepoEntry(entry_name=repo_name,
                     name=plugin_name,
                     python=json_value['python'],
                     repo=json_value['repo'],
                     path=json_value['path'],
                     avatar_url=json_value['avatar_url'],
                     documentation=json_value['documentation'])


def tokenizeJsonEntry(json_dict):
    """
    Returns all the words in a repo entry.
    """
    search = ' '.join((str(word) for word in json_dict.values()))
    return set(FIND_WORDS_RE.findall(search.lower()))


def which(program):
    if ON_WINDOWS:
        program += '.exe'

    def is_exe(file_path):
        return os.path.isfile(file_path) and os.access(file_path, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def check_dependencies(req_path: Path) -> Tuple[Union[str, None], Sequence[str]]:
    """ This methods returns a pair of (message, packages missing).
    Or None, [] if everything is OK.
    """
    log.debug('check dependencies of %s', req_path)
    # noinspection PyBroadException
    try:
        from pkg_resources import get_distribution
        missing_pkg = []

        if not req_path.is_file():
            log.debug('%s has no requirements.txt file', req_path)
            return None, missing_pkg

        with req_path.open() as f:
            for line in f:
                stripped = line.strip()
                # skip empty lines.
                if not stripped:
                    continue

                # noinspection PyBroadException
                try:
                    get_distribution(stripped)
                except Exception:
                    missing_pkg.append(stripped)
        if missing_pkg:
            return f'You need these dependencies for {req_path}: ' + ','.join(missing_pkg), missing_pkg
        return None, missing_pkg
    except Exception:
        log.exception('Problem checking for dependencies.')
        return 'You need to have setuptools installed for the dependency check of the plugins', []


class BotRepoManager(StoreMixin):
    """
    Manages the repo list, git clones/updates or the repos.
    """
    def __init__(self, storage_plugin: StoragePluginBase, plugin_dir: str, plugin_indexes: Tuple[str, ...]) -> None:
        """
        Make a repo manager.
        :param storage_plugin: where the manager store its state.
        :param plugin_dir: where on disk it will git clone the repos.
        :param plugin_indexes: a list of URL / path to get the json repo index.
        """
        super().__init__()
        self.plugin_indexes = plugin_indexes
        self.storage_plugin = storage_plugin
        self.plugin_dir = plugin_dir
        self.open_storage(storage_plugin, 'repomgr')

    def shutdown(self) -> None:
        self.close_storage()

    def check_for_index_update(self) -> None:
        if REPO_INDEX not in self:
            log.info('No repo index, creating it.')
            self.index_update()
            return

        if datetime.fromtimestamp(self[REPO_INDEX][LAST_UPDATE]) < datetime.now() - REPO_INDEXES_CHECK_INTERVAL:
            log.info('Index is too old, update it.')
            self.index_update()

    def index_update(self) -> None:
        index = {LAST_UPDATE: datetime.now().timestamp()}
        for source in reversed(self.plugin_indexes):
            try:
                if urlparse(source).scheme in ('http', 'https'):
                    req = Request(source, headers={'User-Agent': 'Errbot'})
                    with urlopen(url=req, timeout=10) as request:  # nosec
                        log.debug('Update from remote source %s...', source)
                        encoding = request.headers.get_content_charset()
                        content = request.read().decode(encoding if encoding else 'utf-8')
                else:
                    with open(source, encoding='utf-8', mode='r') as src_file:
                        log.debug('Update from local source %s...', source)
                        content = src_file.read()
                index.update(json.loads(content))
            except (HTTPError, URLError, IOError):
                log.exception('Could not update from source %s, keep the index as it is.', source)
                break
        else:
            # nothing failed so ok, we can store the index.
            self[REPO_INDEX] = index
            log.debug('Stored %d repo entries.', len(index) - 1)

    def get_repo_from_index(self, repo_name: str) -> List[RepoEntry]:
        """
        Retrieve the list of plugins for the repo_name from the index.

        :param repo_name: the name of the repo
        :return: a list of RepoEntry
        """
        plugins = self[REPO_INDEX].get(repo_name, None)
        if plugins is None:
            return None
        result = []
        for name, plugin in plugins.items():
            result.append(makeEntry(repo_name, name, plugin))
        return result

    def search_repos(self, query: str) -> Generator[RepoEntry, None, None]:
        """
        A simple search feature, keywords are AND and case insensitive on all the fields.

        :param query: a string query
        :return: an iterator of RepoEntry
        """
        # first see if we are up to date.
        self.check_for_index_update()
        if REPO_INDEX not in self:
            log.error('No index.')
            return
        query_work_set = set(FIND_WORDS_RE.findall(query.lower()))
        for repo_name, plugins in self[REPO_INDEX].items():
            if repo_name == LAST_UPDATE:
                continue
            for plugin_name, plugin in plugins.items():
                if query_work_set.intersection(tokenizeJsonEntry(plugin)):
                    yield makeEntry(repo_name, plugin_name, plugin)

    def get_installed_plugin_repos(self) -> Dict[str, str]:
        return self.get(INSTALLED_REPOS, {})

    def add_plugin_repo(self, name: str, url: str) -> None:
        with self.mutable(INSTALLED_REPOS, {}) as repos:
            repos[name] = url

    def set_plugin_repos(self, repos: Dict[str, str]) -> None:
        """ Used externally.
        """
        self[INSTALLED_REPOS] = repos

    def get_all_repos_paths(self) -> List[str]:
        return [os.path.join(self.plugin_dir, d) for d in self.get(INSTALLED_REPOS, {}).keys()]

    def install_repo(self, repo: str) -> str:
        """
        Install the repository from repo

        :param repo:
            The url, git url or path on disk of a repository. It can point to either a git repo or
             a .tar.gz of a plugin
        :returns:
            The path on disk where the repo has been installed on.
        :raises: :class:`~RepoException` if an error occured.
        """
        self.check_for_index_update()

        human_name = None
        # try to find if we have something with that name in our index
        if repo in self[REPO_INDEX]:
            human_name = repo
            repo_url = next(iter(self[REPO_INDEX][repo].values()))['repo']
        elif not repo.endswith('tar.gz'):
            # This is a repo url, make up a plugin definition for it
            human_name = human_name_for_git_url(repo)
            repo_url = repo
        else:
            repo_url = repo

        # TODO: Update download path of plugin.
        if repo_url.endswith('tar.gz'):
            fo = urlopen(repo_url)  # nosec
            tar = tarfile.open(fileobj=fo, mode='r:gz')
            tar.extractall(path=self.plugin_dir)
            s = repo_url.split(':')[-1].split('/')[-1]
            human_name = s[:-len('.tar.gz')]
        else:
            human_name = human_name or human_name_for_git_url(repo_url)
            try:
                git_clone(repo_url, os.path.join(self.plugin_dir, human_name))
            except Exception as exception:  # dulwich errors all base on exceptions.Exception
                raise RepoException(f'Could not load this plugin: \n\n{repo_url}\n\n---\n\n{exception}')

        self.add_plugin_repo(human_name, repo_url)
        return os.path.join(self.plugin_dir, human_name)

    def update_repos(self, repos) -> Generator[Tuple[str, int, str], None, None]:
        """
        This git pulls the specified repos on disk.
        Yields tuples like (name, success, reason)
        """
        # protects for update outside of what we know is installed
        names = set(self.get_installed_plugin_repos().keys()).intersection(set(repos))

        for d in (path.join(self.plugin_dir, name) for name in names):
            success = 1
            try:
                git_pull(d)
                feedback = "Pulled remote"
                success = 0
            except Exception as exception:
                feedback = f"Error pulling remote {exception}"
                pass

            dep_err, missing_pkgs = check_dependencies(Path(d) / 'requirements.txt')
            if dep_err:
                feedback += dep_err + '\n'
            yield d, success, feedback

    def update_all_repos(self) -> Generator[Tuple[str, int, str], None, None]:
        return self.update_repos(self.get_installed_plugin_repos().keys())

    def uninstall_repo(self, name: str) -> None:
        repo_path = path.join(self.plugin_dir, name)
        # ignore errors because the DB can be desync'ed from the file tree.
        shutil.rmtree(repo_path, ignore_errors=True)
        repos = self.get_installed_plugin_repos()
        del(repos[name])
        self.set_plugin_repos(repos)



import collections
import fnmatch
import inspect
import logging
import os
import re
import sys
import time
from platform import system
from functools import wraps
from typing import List


from dulwich import porcelain

log = logging.getLogger(__name__)

ON_WINDOWS = system() == 'Windows'

PLUGINS_SUBDIR = 'plugins'


# noinspection PyPep8Naming
class deprecated(object):
    """ deprecated decorator. emits a warning on a call on an old method and call the new method anyway """
    def __init__(self, new=None):
        self.new = new

    def __call__(self, old):
        @wraps(old)
        def wrapper(*args, **kwds):
            frame = inspect.getframeinfo(inspect.currentframe().f_back)
            msg = f'{frame.filename}: {frame.lineno}: '
            if len(args):
                pref = type(args[0]).__name__ + '.'  # TODO might break for individual methods
            else:
                pref = ''
            msg += f'call to the deprecated {pref}{old.__name__}'
            if self.new is not None:
                if type(self.new) is property:
                    msg += f'... use the property {pref}{self.new.fget.__name__} instead'
                else:
                    msg += f'... use {pref}{self.new.__name__} instead'
            msg += '.'
            logging.warning(msg)

            if self.new:
                if type(self.new) is property:
                    return self.new.fget(*args, **kwds)
                return self.new(*args, **kwds)
            return old(*args, **kwds)

        wrapper.__name__ = old.__name__
        wrapper.__doc__ = old.__doc__
        wrapper.__dict__.update(old.__dict__)
        return wrapper


def format_timedelta(timedelta):
    total_seconds = timedelta.seconds + (86400 * timedelta.days)
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours == 0 and minutes == 0:
        return f'{seconds:d} seconds'
    elif not hours:
        return f'{minutes:d} minutes'
    elif not minutes:
        return f'{hours:d} hours'
    return f'{hours:d} hours and {minutes:d} minutes'


INVALID_VERSION_EXCEPTION = 'version %s in not in format "x.y.z" or "x.y.z-{beta,alpha,rc1,rc2...}" for example "1.2.2"'


def version2tuple(version):
    vsplit = version.split('-')

    if len(vsplit) == 2:
        main, sub = vsplit
        if sub == 'alpha':
            sub_int = -1
        elif sub == 'beta':
            sub_int = 0
        elif sub.startswith('rc'):
            sub_int = int(sub[2:])
        else:
            raise ValueError(INVALID_VERSION_EXCEPTION % version)

    elif len(vsplit) == 1:
        main = vsplit[0]
        sub_int = sys.maxsize
    else:
        raise ValueError(INVALID_VERSION_EXCEPTION % version)

    response = [int(el) for el in main.split('.')]
    response.append(sub_int)

    if len(response) != 4:
        raise ValueError(INVALID_VERSION_EXCEPTION % version)

    return tuple(response)


REMOVE_EOL = re.compile(r'\n')
REINSERT_EOLS = re.compile(r'</p>|</li>|<br/>', re.I)
ZAP_TAGS = re.compile(r'<[^>]+>')


def rate_limited(min_interval):
    """
    decorator to rate limit a function.

    :param min_interval: minimum interval allowed between 2 consecutive calls.
    :return: the decorated function
    """
    def decorate(func):
        last_time_called = [0.0]

        def rate_limited_function(*args, **kargs):
            elapsed = time.time() - last_time_called[0]
            log.debug('Elapsed %f since last call', elapsed)
            left_to_wait = min_interval - elapsed
            if left_to_wait > 0:
                log.debug('Wait %f due to rate limiting...', left_to_wait)
                time.sleep(left_to_wait)
            ret = func(*args, **kargs)
            last_time_called[0] = time.time()
            return ret

        return rate_limited_function

    return decorate


def split_string_after(str_, n):
    """Yield chunks of length `n` from the given string

    :param n: length of the chunks.
    :param str_: the given string.
    """
    for start in range(0, max(len(str_), 1), n):
        yield str_[start:start + n]


def find_roots(path: str, file_sig: str = '*.plug') -> List:
    """Collects all the paths from path recursively that contains files of type `file_sig`.

       :param path:
            a base path to walk from
       :param file_sig:
            the file pattern to look for
       :return: a list of paths
    """
    roots = list()  # you can have several .plug per directory.
    for root, dirnames, filenames in os.walk(path, followlinks=True):
        for filename in fnmatch.filter(filenames, file_sig):
            dir_to_add = os.path.dirname(os.path.join(root, filename))
            relative = os.path.relpath(os.path.realpath(dir_to_add), os.path.realpath(path))
            for subelement in relative.split(os.path.sep):
                # if one of the element is just a relative construct, it is ok to continue inspecting it.
                if subelement in ('.', '..'):
                    continue
                # if it is an hidden directory or a python temp directory, just ignore it.
                if subelement.startswith('.') or subelement == '__pycache__':
                    log.debug('Ignore %s.', dir_to_add)
                    break
            else:
                roots.append(dir_to_add)
    return list(collections.OrderedDict.fromkeys(roots))


def collect_roots(base_paths: List, file_sig: str = '*.plug') -> List:
    """Collects all the paths from base_paths recursively that contains files of type `file_sig`.

       :param base_paths:
            a list of base paths to walk from
            elements can be a string or a list/tuple of strings

       :param file_sig:
            the file pattern to look for
       :return: a list of paths
    """
    result = list()
    for path_or_list in base_paths:
        if isinstance(path_or_list, (list, tuple)):
            result.extend(collect_roots(base_paths=path_or_list, file_sig=file_sig))
        elif path_or_list is not None:
            result.extend(find_roots(path_or_list, file_sig))
    return list(collections.OrderedDict.fromkeys(result))


def global_restart():
    """Restart the current process."""
    python = sys.executable
    os.execl(python, python, *sys.argv)


def git_clone(url: str, path: str) -> None:
    """
    Clones a repository from git url to path
    """
    if not os.path.exists(path):
        os.makedirs(path)

    repo = porcelain.clone(url, path)
    config = repo.get_config()
    config.set(('branch', 'master'), 'remote', 'origin')
    config.set(('branch', 'master'), 'merge', 'refs/heads/master')
    config.write_to_path()


def git_pull(repo_path: str) -> None:
    """
    Does a git pull on a repository
    """
    porcelain.pull(repo_path)


def git_tag_list(repo_path: str) -> List[str]:
    """
    Lists git tags on a cloned repo
    """
    porcelain.tag_list(repo_path)



import sys
import os
from json import loads
from random import randrange
from threading import Thread

from webtest import TestApp
from errbot.core_plugins import flask_app
from werkzeug.serving import ThreadedWSGIServer

from errbot import botcmd, BotPlugin, webhook

from urllib.request import unquote

from OpenSSL import crypto

TEST_REPORT = """*** Test Report
URL : %s
Detected your post as : %s
Status code : %i
"""


def make_ssl_certificate(key_path, cert_path):
    """
    Generate a self-signed certificate

    The generated key will be written out to key_path, with the corresponding
    certificate itself being written to cert_path.
    :param cert_path: path where to write the certificate.
    :param key_path: path where to write the key.
    """
    cert = crypto.X509()
    cert.set_serial_number(randrange(1, sys.maxsize))
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(60 * 60 * 24 * 365)

    subject = cert.get_subject()
    subject.CN = '*'
    setattr(subject, 'O', 'Self-Signed Certificate for Errbot')  # Pep8 annoyance workaround

    issuer = cert.get_issuer()
    issuer.CN = 'Self-proclaimed Authority'
    setattr(issuer, 'O', 'Self-Signed')  # Pep8 annoyance workaround

    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 4096)
    cert.set_pubkey(pkey)
    cert.sign(pkey, 'sha256')

    f = open(cert_path, 'w')
    f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode('utf-8'))
    f.close()

    f = open(key_path, 'w')
    f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey).decode('utf-8'))
    f.close()


class Webserver(BotPlugin):

    def __init__(self, *args, **kwargs):
        self.server = None
        self.server_thread = None
        self.ssl_context = None
        self.test_app = TestApp(flask_app)
        super().__init__(*args, **kwargs)

    def get_configuration_template(self):
        return {'HOST': '0.0.0.0',
                'PORT': 3141,
                'SSL': {'enabled': False,
                        'host': '0.0.0.0',
                        'port': 3142,
                        'certificate': '',
                        'key': ''}}

    def check_configuration(self, configuration):
        # it is a pain, just assume a default config if SSL is absent or set to None
        if configuration.get('SSL', None) is None:
            configuration['SSL'] = {'enabled': False, 'host': '0.0.0.0', 'port': 3142, 'certificate': '', 'key': ''}
        super().check_configuration(configuration)

    def activate(self):
        if not self.config:
            self.log.info('Webserver is not configured. Forbid activation')
            return

        if self.server_thread and self.server_thread.is_alive():
            raise Exception('Invalid state, you should not have a webserver already running.')
        self.server_thread = Thread(target=self.run_server, name='Webserver Thread')
        self.server_thread.start()
        self.log.debug('Webserver started.')

        super().activate()

    def deactivate(self):
        if self.server is not None:
            self.log.info('Shutting down the internal webserver.')
            self.server.shutdown()
            self.log.info('Waiting for the webserver thread to quit.')
            self.server_thread.join()
            self.log.info('Webserver shut down correctly.')
        super().deactivate()

    def run_server(self):
        try:
            host = self.config['HOST']
            port = self.config['PORT']
            ssl = self.config['SSL']
            self.log.info('Starting the webserver on %s:%i', host, port)
            ssl_context = (ssl['certificate'], ssl['key']) if ssl['enabled'] else None
            self.server = ThreadedWSGIServer(host, ssl['port'] if ssl_context else port, flask_app,
                                             ssl_context=ssl_context)
            self.server.serve_forever()
            self.log.debug('Webserver stopped')
        except KeyboardInterrupt:
            self.log.info('Keyboard interrupt, request a global shutdown.')
            self.server.shutdown()
        except Exception:
            self.log.exception('The webserver exploded.')

    @botcmd(template='webstatus')
    def webstatus(self, msg, args):
        """
        Gives a quick status of what is mapped in the internal webserver
        """
        return {'rules': (((rule.rule, rule.endpoint) for rule in flask_app.url_map._rules))}

    @webhook
    def echo(self, incoming_request):
        """
        A simple test webhook
        """
        self.log.debug("Your incoming request is :" + str(incoming_request))
        return str(incoming_request)

    @botcmd(split_args_with=' ')
    def webhook_test(self, _, args):
        """
            Test your webhooks from within err.

        The syntax is :
        !webhook test [relative_url] [post content]

        It triggers the notification and generate also a little test report.
        """
        url = args[0]
        content = ' '.join(args[1:])

        # try to guess the content-type of what has been passed
        try:
            # try if it is plain json
            loads(content)
            contenttype = 'application/json'
        except ValueError:
            # try if it is a form
            splitted = content.split('=')
            # noinspection PyBroadException
            try:
                payload = '='.join(splitted[1:])
                loads(unquote(payload))
                contenttype = 'application/x-www-form-urlencoded'
            except Exception as _:
                contenttype = 'text/plain'  # dunno what it is

        self.log.debug('Detected your post as : %s.', contenttype)

        response = self.test_app.post(url, params=content, content_type=contenttype)
        return TEST_REPORT % (url, contenttype, response.status_code)

    @botcmd(admin_only=True)
    def generate_certificate(self, _, args):
        """
        Generate a self-signed SSL certificate for the Webserver
        """
        yield ('Generating a new private key and certificate. This could take a '
               'while if your system is slow or low on entropy')
        key_path = os.sep.join((self.bot_config.BOT_DATA_DIR, "webserver_key.pem"))
        cert_path = os.sep.join((self.bot_config.BOT_DATA_DIR, "webserver_certificate.pem"))
        make_ssl_certificate(key_path=key_path, cert_path=cert_path)
        yield f'Certificate successfully generated and saved in {self.bot_config.BOT_DATA_DIR}.'

        suggested_config = self.config
        suggested_config['SSL']['enabled'] = True
        suggested_config['SSL']['host'] = suggested_config['HOST']
        suggested_config['SSL']['port'] = suggested_config['PORT'] + 1
        suggested_config['SSL']['key'] = key_path
        suggested_config['SSL']['certificate'] = cert_path
        yield 'To enable SSL with this certificate, the following config is recommended:'
        yield f'{suggested_config!r}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Errbot documentation build configuration file, created by
# sphinx-quickstart on Fri Sep 13 17:24:59 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import subprocess, sys, os
sys.path.append(os.path.abspath('_themes'))
sys.path.append(os.path.abspath('_themes/err'))
sys.path.append(os.path.abspath('../'))

__import__('errbot.config-template')
sys.modules['config'] = sys.modules['errbot.config-template']
from errbot.version import VERSION

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_annotation',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Err'
copyright = '2013, Guillaume Binet, Tali Davidovich Petrover and Nick Groenen'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = VERSION
# The full version, including alpha/beta/rc tags.
release = VERSION

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    '_build',
    'error_pages/*',
    'errbot.backends.campfire.rst',  # Broken on Python 3
    'errbot.backends.graphic.rst',  # Quite a pain to build, not worth it
    'errbot.backends.tox.rst',  # Also quite a pain to build at this stage
    '_gh-pages',
]

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# -- Autodoc configuration -----------------------------------------------------


def autodoc_skip_member(app, what, name, obj, skip, options):
    if name == "__init__":
        return False
    return skip


# -- Apidoc --------------------------------------------------------------------


def run_apidoc(_):
    subprocess.check_call("sphinx-apidoc --separate -f -o . ../errbot", shell=True)


def run_repos_builder(*_):
    last_dir = os.getcwd()
    os.chdir('../tools/')
    subprocess.check_call("python plugin-gen.py", shell=True, env=os.environ)
    subprocess.check_call("pwd; cp repos.json ../docs/html_extras/", shell=True)
    os.chdir(last_dir)


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'err'

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_themes']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#
# Note: Best leave this off until Sphinx issue #1496 gets resolved
# (https://bitbucket.org/birkenfeld/sphinx/issue/1496/smartypants-should-not-convert-text-in-a)
html_use_smartypants = False

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Errdoc'

html_extra_path = ['html_extras']

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ('index', 'Err.tex', 'Err Documentation', 'Guillaume Binet, Tali Davidovich Petrover and Nick Groenen', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', 'err', 'Err Documentation', ['Guillaume Binet, Tali Davidovich Petrover and Nick Groenen'], 1)]

# If true, show URL addresses after external links.
#man_show_urls = False

# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'Err', 'Err Documentation', 'Guillaume Binet, Tali Davidovich Petrover and Nick Groenen', 'Err',
     'One line description of project.', 'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = 'Err'
epub_author = 'Guillaume Binet, Tali Davidovich Petrover and Nick Groenen'
epub_publisher = 'Guillaume Binet, Tali Davidovich Petrover and Nick Groenen'
epub_copyright = '2013, Guillaume Binet, Tali Davidovich Petrover and Nick Groenen'

# The language of the text. It defaults to the language option
# or en if the language is not set.
#epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
#epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#epub_identifier = ''

# A unique identification for the text.
#epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
#epub_cover = ()

# A sequence of (type, uri, title) tuples for the guide element of content.opf.
#epub_guide = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_post_files = []

# A list of files that should not be packed into the epub file.
#epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
#epub_tocdepth = 3

# Allow duplicate toc entries.
#epub_tocdup = True

# Fix unsupported image types using the PIL.
#epub_fix_images = False

# Scale large images.
#epub_max_image_width = 0

# If 'no', URL addresses will not be shown.
#epub_show_urls = 'inline'

# If false, no index is generated.
#epub_use_index = True

# -- Misc options -------------------------------------------------------------

intersphinx_mapping = {'http://docs.python.org/': None}


def setup(app):
    app.connect("autodoc-skip-member", autodoc_skip_member)
    app.connect("builder-inited", run_apidoc)
    #app.connect("html-page-context", run_repos_builder)



import importlib
import logging
import sys
import unittest
import textwrap
from os.path import sep, abspath
from queue import Queue, Empty
from tempfile import mkdtemp
from threading import Thread

import pytest

from errbot.rendering import text
from errbot.backends.base import Message, Room, Person, RoomOccupant, ONLINE
from errbot.core_plugins.wsview import reset_app
from errbot.core import ErrBot
from errbot.bootstrap import setup_bot
from errbot.utils import deprecated

log = logging.getLogger(__name__)

QUIT_MESSAGE = '$STOP$'

STZ_MSG = 1
STZ_PRE = 2
STZ_IQ = 3


class TestPerson(Person):
    """
    This is an identifier just represented as a string.
    DO NOT USE THIS DIRECTLY AS IT IS NOT COMPATIBLE WITH MOST BACKENDS,
    use self.build_identifier(identifier_as_string) instead.

    Note to back-end implementors: You should provide a custom
    <yourbackend>Identifier object that adheres to this interface.

    You should not directly inherit from SimpleIdentifier, inherit
    from object instead and make sure it includes all properties and
    methods exposed by this class.
    """

    def __init__(self, person, client=None, nick=None, fullname=None):
        self._person = person
        self._client = client
        self._nick = nick
        self._fullname = fullname

    @property
    def person(self):
        """This needs to return the part of the identifier pointing to a person."""
        return self._person

    @property
    def client(self):
        """This needs to return the part of the identifier pointing to a client
        from which a person is sending a message from.
        Returns None is unspecified"""
        return self._client

    @property
    def nick(self):
        """This needs to return a short display name for this identifier e.g. gbin.
        Returns None is unspecified"""
        return self._nick

    @property
    def fullname(self):
        """This needs to return a long display name for this identifier e.g. Guillaume Binet.
        Returns None is unspecified"""
        return self._fullname

    aclattr = person

    def __unicode__(self):
        if self.client:
            return f'{self._person}/{self._client}'
        return f'{self._person}'

    __str__ = __unicode__

    def __eq__(self, other):
        if not isinstance(other, Person):
            return False
        return self.person == other.person


# noinspection PyAbstractClass
class TestOccupant(TestPerson, RoomOccupant):
    """ This is a MUC occupant represented as a string.
        DO NOT USE THIS DIRECTLY AS IT IS NOT COMPATIBLE WITH MOST BACKENDS,
    """

    def __init__(self, person, room):
        super().__init__(person)
        self._room = room

    @property
    def room(self):
        return self._room

    def __unicode__(self):
        return self._person + '@' + str(self._room)

    __str__ = __unicode__

    def __eq__(self, other):
        return self.person == other.person and self.room == other.room


class TestRoom(Room):
    def invite(self, *args):
        pass

    def __init__(self, name, occupants=None, topic=None, bot=None):
        """
        :param name: Name of the room
        :param occupants: Occupants of the room
        :param topic: The MUC's topic
        """
        if occupants is None:
            occupants = []
        self._occupants = occupants
        self._topic = topic
        self._bot = bot
        self._name = name
        self._bot_mucid = TestOccupant(self._bot.bot_config.BOT_IDENTITY['username'], self._name)

    @property
    def occupants(self):
        return self._occupants

    def find_croom(self):
        """ find back the canonical room from a this room"""
        for croom in self._bot._rooms:
            if croom == self:
                return croom
        return None

    @property
    def joined(self):
        room = self.find_croom()
        if room:
            return self._bot_mucid in room.occupants
        return False

    def join(self, username=None, password=None):
        if self.joined:
            logging.warning('Attempted to join room %s, but already in this room.', self)
            return

        if not self.exists:
            log.debug("Room %s doesn't exist yet, creating it.", self)
            self.create()

        room = self.find_croom()
        room._occupants.append(self._bot_mucid)
        log.info('Joined room %s.', self)
        self._bot.callback_room_joined(room)

    def leave(self, reason=None):
        if not self.joined:
            logging.warning('Attempted to leave room %s, but not in this room.', self)
            return

        room = self.find_croom()
        room._occupants.remove(self._bot_mucid)
        log.info('Left room %s.', self)
        self._bot.callback_room_left(room)

    @property
    def exists(self):
        return self.find_croom() is not None

    def create(self):
        if self.exists:
            logging.warning('Room %s already created.', self)
            return

        self._bot._rooms.append(self)
        log.info('Created room %s.', self)

    def destroy(self):
        if not self.exists:
            logging.warning("Cannot destroy room %s, it doesn't exist.", self)
            return

        self._bot._rooms.remove(self)
        log.info('Destroyed room %s.', self)

    @property
    def topic(self):
        return self._topic

    @topic.setter
    def topic(self, topic):
        self._topic = topic
        room = self.find_croom()
        room._topic = self._topic
        log.info('Topic for room %s set to %s.', self, topic)
        self._bot.callback_room_topic(self)

    def __unicode__(self):
        return self._name

    def __str__(self):
        return self._name

    def __eq__(self, other):
        return self._name == other._name


class TestBackend(ErrBot):
    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        pass

    def __init__(self, config):
        config.BOT_LOG_LEVEL = logging.DEBUG
        config.CHATROOM_PRESENCE = ('testroom',)  # we are testing with simple identfiers
        config.BOT_IDENTITY = {'username': 'err'}  # we are testing with simple identfiers
        self.bot_identifier = self.build_identifier('Err')  # whatever

        super().__init__(config)
        self.incoming_stanza_queue = Queue()
        self.outgoing_message_queue = Queue()
        self.sender = self.build_identifier(config.BOT_ADMINS[0])  # By default, assume this is the admin talking
        self.reset_rooms()
        self.md = text()

    def send_message(self, msg):
        log.info("\n\n\nMESSAGE:\n%s\n\n\n", msg.body)
        super().send_message(msg)
        self.outgoing_message_queue.put(self.md.convert(msg.body))

    def send_stream_request(self, user, fsource, name, size, stream_type):
        # Just dump the stream contents to the message queue
        self.outgoing_message_queue.put(fsource.read())

    def serve_forever(self):
        self.connect_callback()  # notify that the connection occured
        try:
            while True:
                log.debug('waiting on queue')
                stanza_type, entry = self.incoming_stanza_queue.get()
                log.debug('message received')
                if entry == QUIT_MESSAGE:
                    log.info("Stop magic message received, quitting...")
                    break
                if stanza_type is STZ_MSG:
                    msg = Message(entry)
                    msg.frm = self.sender
                    msg.to = self.bot_identifier  # To me only

                    self.callback_message(msg)

                    # implements the mentions.
                    mentioned = [self.build_identifier(word[1:]) for word in entry.split() if word.startswith('@')]
                    if mentioned:
                        self.callback_mention(msg, mentioned)

                elif stanza_type is STZ_PRE:
                    log.info("Presence stanza received.")
                    self.callback_presence(entry)
                elif stanza_type is STZ_IQ:
                    log.info("IQ stanza received.")
                else:
                    log.error("Unknown stanza type.")

        except EOFError:
            pass
        except KeyboardInterrupt:
            pass
        finally:
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    def connect(self):
        return

    def build_identifier(self, text_representation):
        return TestPerson(text_representation)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        msg = self.build_message(text)
        msg.frm = self.bot_identifier
        msg.to = msg.frm
        return msg

    @property
    def mode(self):
        return 'test'

    def rooms(self):
        return [r for r in self._rooms if r.joined]

    def query_room(self, room):
        try:
            return [r for r in self._rooms if str(r) == str(room)][0]
        except IndexError:
            r = TestRoom(room, bot=self)
            return r

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'@{identifier.nick} {message.body}'

    def pop_message(self, timeout=5, block=True):
        return self.outgoing_message_queue.get(timeout=timeout, block=block)

    def push_message(self, msg):
        self.incoming_stanza_queue.put((STZ_MSG, msg), timeout=5)

    def push_presence(self, presence):
        """ presence must at least duck type base.Presence
        """
        self.incoming_stanza_queue.put((STZ_PRE, presence), timeout=5)

    def zap_queues(self):
        while not self.incoming_stanza_queue.empty():
            msg = self.incoming_stanza_queue.get(block=False)
            log.error('Message left in the incoming queue during a test: %s.', msg)

        while not self.outgoing_message_queue.empty():
            msg = self.outgoing_message_queue.get(block=False)
            log.error('Message left in the outgoing queue during a test: %s.', msg)

    def reset_rooms(self):
        """Reset/clear all rooms"""
        self._rooms = []


class ShallowConfig(object):
    pass


class TestBot(object):
    """
    A minimal bot utilizing the TestBackend, for use with unit testing.

    Only one instance of this class should globally be active at any one
    time.

    End-users should not use this class directly. Use
    :func:`~errbot.backends.test.testbot` or
    :class:`~errbot.backends.test.FullStackTest` instead, which use this
    class under the hood.
    """
    bot_thread = None

    def __init__(self, extra_plugin_dir=None, loglevel=logging.DEBUG, extra_config=None):
        self.setup(extra_plugin_dir=extra_plugin_dir, loglevel=loglevel, extra_config=extra_config)

    def setup(self, extra_plugin_dir=None, loglevel=logging.DEBUG, extra_config=None):
        """
        :param extra_config: Piece of extra configuration you want to inject to the config.
        :param extra_plugin_dir: Path to a directory from which additional
            plugins should be loaded.
        :param loglevel: Logging verbosity. Expects one of the constants
            defined by the logging module.
        """
        tempdir = mkdtemp()

        # This is for test isolation.
        config = ShallowConfig()
        config.__dict__.update(importlib.import_module('errbot.config-template').__dict__)
        config.BOT_DATA_DIR = tempdir
        config.BOT_LOG_FILE = tempdir + sep + 'log.txt'
        config.STORAGE = 'Memory'

        if extra_config is not None:
            log.debug('Merging %s to the bot config.', repr(extra_config))
            for k, v in extra_config.items():
                setattr(config, k, v)

        # reset logging to console
        logging.basicConfig(format='%(levelname)s:%(message)s')
        file = logging.FileHandler(config.BOT_LOG_FILE, encoding='utf-8')
        self.logger = logging.getLogger('')
        self.logger.setLevel(loglevel)
        self.logger.addHandler(file)

        config.BOT_EXTRA_PLUGIN_DIR = extra_plugin_dir
        config.BOT_LOG_LEVEL = loglevel
        self.bot_config = config

    def start(self, timeout=2):
        """
        Start the bot

        Calling this method when the bot has already started will result
        in an Exception being raised.
        :param timeout: Timeout for the ready message pop. pop will be done 60 times so the total timeout is 60*timeout
        """
        if self.bot_thread is not None:
            raise Exception("Bot has already been started")
        self._bot = setup_bot('Test', self.logger, self.bot_config)
        self.bot_thread = Thread(target=self.bot.serve_forever, name='TestBot main thread')
        self.bot_thread.setDaemon(True)
        self.bot_thread.start()

        self.bot.push_message("!echo ready")

        # Ensure bot is fully started and plugins are loaded before returning
        try:
            for i in range(60):
                #  Gobble initial error messages...
                msg = self.bot.pop_message(timeout=timeout)
                if msg == "ready":
                    break
                log.warning("Queue was not empty, the non-consumed message is:")
                log.warning(msg)
                log.warning("Check the previous test and remove spurrious messages.")
        except Empty:
            raise AssertionError('The "ready" message has not been received (timeout).')

    @property
    def bot(self) -> ErrBot:
        return self._bot

    def stop(self):
        """
        Stop the bot

        Calling this method before the bot has started will result in an
        Exception being raised.
        """
        if self.bot_thread is None:
            raise Exception("Bot has not yet been started")
        self.bot.push_message(QUIT_MESSAGE)
        self.bot_thread.join()
        reset_app()  # empty the bottle ... hips!
        log.info("Main bot thread quits")
        self.bot.zap_queues()
        self.bot.reset_rooms()
        self.bot_thread = None

    def pop_message(self, timeout=5, block=True):
        return self.bot.pop_message(timeout, block)

    def push_message(self, msg):
        return self.bot.push_message(msg)

    def push_presence(self, presence):
        """ presence must at least duck type base.Presence
        """
        return self.bot.push_presence(presence)

    def exec_command(self, command, timeout=5):
        """ Execute a command and return the first response.
        This makes more py.test'ist like:
        assert 'blah' in exec_command('!hello')
        """
        self.bot.push_message(command)
        return self.bot.pop_message(timeout)

    def zap_queues(self):
        return self.bot.zap_queues()

    def assertInCommand(self, command, response, timeout=5, dedent=False):
        """Assert the given command returns the given response"""
        if dedent:
            command = '\n'.join(textwrap.dedent(command).splitlines()[1:])
        self.bot.push_message(command)
        msg = self.bot.pop_message(timeout)
        assert response in msg, f'{response} not in {msg}.'

    @deprecated(assertInCommand)
    def assertCommand(self, command, response, timeout=5, dedent=False):
        """Assert the given command returns the given response"""
        pass

    def assertCommandFound(self, command, timeout=5):
        """Assert the given command exists"""
        self.bot.push_message(command)
        assert 'not found' not in self.bot.pop_message(timeout)

    def inject_mocks(self, plugin_name: str, mock_dict: dict):
        """Inject mock objects into the plugin

        mock_dict = {
            'field_1': obj_1,
            'field_2': obj_2,
        }
        testbot.inject_mocks(HelloWorld, mock_dict)
        assert 'blah' in testbot.exec_command('!hello')
        """
        plugin = self.bot.plugin_manager.get_plugin_obj_by_name(plugin_name)

        if plugin is None:
            raise Exception(f'"{plugin_name}" is not loaded.')
        for field, mock_obj in mock_dict.items():
            if not hasattr(plugin, field):
                raise ValueError(f'No property/attribute named "{field}" attached.')
            setattr(plugin, field, mock_obj)


class FullStackTest(unittest.TestCase, TestBot):
    """
    Test class for use with Python's unittest module to write tests
    against a fully functioning bot.

    For example, if you wanted to test the builtin `!about` command,
    you could write a test file with the following::

        from errbot.backends.test import FullStackTest

        class TestCommands(FullStackTest):
            def test_about(self):
                self.push_message('!about')
                self.assertIn('Err version', self.pop_message())
    """

    def setUp(self, extra_plugin_dir=None, extra_test_file=None, loglevel=logging.DEBUG, extra_config=None):
        """
        :param extra_plugin_dir: Path to a directory from which additional
            plugins should be loaded.
        :param extra_test_file: [Deprecated but kept for backward-compatibility,
            use extra_plugin_dir instead]
            Path to an additional plugin which should be loaded.
        :param loglevel: Logging verbosity. Expects one of the constants
            defined by the logging module.
        :param extra_config: Piece of extra bot config in a dict.
        """
        if extra_plugin_dir is None and extra_test_file is not None:
            extra_plugin_dir = sep.join(abspath(extra_test_file).split(sep)[:-2])

        self.setup(extra_plugin_dir=extra_plugin_dir, loglevel=loglevel, extra_config=extra_config)
        self.start()

    def tearDown(self):
        self.stop()


@pytest.fixture
def testbot(request) -> TestBot:
    """
    Pytest fixture to write tests against a fully functioning bot.

    For example, if you wanted to test the builtin `!about` command,
    you could write a test file with the following::

        def test_about(testbot):
            testbot.push_message('!about')
            assert "Err version" in testbot.pop_message()

    It's possible to provide additional configuration to this fixture,
    by setting variables at module level or as class attributes (the
    latter taking precedence over the former). For example::

        extra_plugin_dir = '/foo/bar'

        def test_about(testbot):
            testbot.push_message('!about')
            assert "Err version" in testbot.pop_message()

    ..or::

        extra_plugin_dir = '/foo/bar'

        class Tests(object):
            # Wins over `extra_plugin_dir = '/foo/bar'` above
            extra_plugin_dir = '/foo/baz'

            def test_about(self, testbot):
                testbot.push_message('!about')
                assert "Err version" in testbot.pop_message()

    ..to load additional plugins from the directory `/foo/bar` or
    `/foo/baz` respectively. This works for the following items, which are
    passed to the constructor of :class:`~errbot.backends.test.TestBot`:

    * `extra_plugin_dir`
    * `loglevel`
    """

    def on_finish():
        bot.stop()

    #  setup the logging to something digestable.
    logger = logging.getLogger('')
    logging.getLogger('MARKDOWN').setLevel(logging.ERROR)  # this one is way too verbose in debug
    logger.setLevel(logging.DEBUG)
    console_hdlr = logging.StreamHandler(sys.stdout)
    console_hdlr.setFormatter(logging.Formatter("%(levelname)-8s %(name)-25s %(message)s"))
    logger.handlers = []
    logger.addHandler(console_hdlr)

    kwargs = {}

    for attr, default in (('extra_plugin_dir', None), ('extra_config', None), ('loglevel', logging.DEBUG),):
        if hasattr(request, 'instance'):
            kwargs[attr] = getattr(request.instance, attr, None)
        if kwargs[attr] is None:
            kwargs[attr] = getattr(request.module, attr, default)

    bot = TestBot(**kwargs)
    bot.start()

    request.addfinalizer(on_finish)
    return bot



from errbot import BotPlugin, ValidationException


class FailP(BotPlugin):
    """
    Just a plugin failing at config time.
    """
    def get_configuration_template(self):
        return {'One': 1, 'Two': 2}

    def check_configuration(self, configuration):
        raise ValidationException('Message explaining why it failed.')



from abc import abstractmethod
from typing import Any, Iterable


class StorageBase(object):
    """
    Contract to implemement a storage.
    """
    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """
        Atomically set the key to the given value.
        The caller of set will protect against set on non open.

        :param key: string as key
        :param value: pickalable python object
        """
        pass

    @abstractmethod
    def get(self, key: str) -> Any:
        """
        Get the value stored for key. Raises KeyError if the key doesn't exist.
        The caller of get will protect against get on non open.

        :param key: the key
        :return: the value
        """
        pass

    @abstractmethod
    def remove(self, key: str) -> None:
        """
        Remove key. Raises KeyError if the key doesn't exist.
        The caller of get will protect against get on non open.

        :param key: the key
        """
        pass

    @abstractmethod
    def len(self) -> int:
        """

        :return: the number of keys set.
        """
        pass

    @abstractmethod
    def keys(self) -> Iterable[str]:
        """

        :return: an iterator on all the entries
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Sync and close the storage.
        The caller of close will protect against close on non open and double close.
        """
        pass


class StoragePluginBase(object):
    """
    Base to implement a storage plugin.
    This is a factory for the namespaces.
    """
    def __init__(self, bot_config):
        self._storage_config = getattr(bot_config, 'STORAGE_CONFIG', {})

    @abstractmethod
    def open(self, namespace: str) -> StorageBase:
        """
        Open the storage with the given namespace (core, or plugin name) and config.
        The caller of open will protect against double opens.

        :param namespace: a namespace to isolate the plugin storages.
        :return:
        """
        pass



from __future__ import absolute_import
from errbot import BotPlugin, botcmd, re_botcmd, botmatch


class DummyTest(BotPlugin):
    """Just a test plugin to see if it is picked up.
    """
    @botcmd
    def foo(self, msg, args):
        """This runs foo."""
        return 'bar'

    @re_botcmd(pattern=r"plz dont match this")
    def re_foo(self, msg, match):
        """This runs re_foo."""
        return 'bar'

    @botmatch(r"match this")
    def re_bar(self, msg, match):
        """This runs re_foo."""
        return 'bar'

    @botcmd
    def run_subcommands(self, msg, args):
        """Tests a simple subcommand"""
        return args

    @botcmd
    def run_lots_of_subcommands(self, msg, args):
        """Tests multiple subcommands"""
        return args

    def helper_method(self, arg):
        return arg

    @botcmd
    def baz(self, msg, args):
        """Tests mock injection method"""
        return self.helper_method('baz')

    @botcmd
    def bar(self, msg, args):
        """This runs bar."""
        return msg



from __future__ import absolute_import
import logging
import sys
import threading
import subprocess
import struct
import re

from markdown import Markdown
from markdown.extensions.extra import ExtraExtension

from errbot.backends.base import Message, Room, RoomError, \
    RoomNotJoinedError, Stream, \
    RoomOccupant, ONLINE, Person
from errbot.core import ErrBot
from errbot.utils import rate_limited
from errbot.rendering.ansiext import AnsiExtension, enable_format, \
    CharacterTable, NSC

log = logging.getLogger(__name__)

IRC_CHRS = CharacterTable(fg_black=NSC('\x0301'),
                          fg_red=NSC('\x0304'),
                          fg_green=NSC('\x0303'),
                          fg_yellow=NSC('\x0308'),
                          fg_blue=NSC('\x0302'),
                          fg_magenta=NSC('\x0306'),
                          fg_cyan=NSC('\x0310'),
                          fg_white=NSC('\x0300'),
                          fg_default=NSC('\x03'),
                          bg_black=NSC('\x03,01'),
                          bg_red=NSC('\x03,04'),
                          bg_green=NSC('\x03,03'),
                          bg_yellow=NSC('\x03,08'),
                          bg_blue=NSC('\x03,02'),
                          bg_magenta=NSC('\x03,06'),
                          bg_cyan=NSC('\x03,10'),
                          bg_white=NSC('\x03,00'),
                          bg_default=NSC('\x03,'),
                          fx_reset=NSC('\x03'),
                          fx_bold=NSC('\x02'),
                          fx_italic=NSC('\x1D'),
                          fx_underline=NSC('\x1F'),
                          fx_not_italic=NSC('\x0F'),
                          fx_not_underline=NSC('\x0F'),
                          fx_normal=NSC('\x0F'),
                          fixed_width='',
                          end_fixed_width='',
                          inline_code='',
                          end_inline_code='')

IRC_NICK_REGEX = r'[a-zA-Z\[\]\\`_\^\{\|\}][a-zA-Z0-9\[\]\\`_\^\{\|\}-]+'
IRC_MESSAGE_SIZE_LIMIT = 510

try:
    import irc.connection
    from irc.client import ServerNotConnectedError, NickMask
    from irc.bot import SingleServerIRCBot
except ImportError:
    log.fatal("""You need the IRC support to use IRC, you can install it with:
    pip install errbot[IRC]
    """)
    sys.exit(-1)


def irc_md():
    """This makes a converter from markdown to mirc color format.
    """
    md = Markdown(output_format='irc', extensions=[ExtraExtension(), AnsiExtension()])
    md.stripTopLevelTags = False
    return md


class IRCPerson(Person):

    def __init__(self, mask):
        self._nickmask = NickMask(mask)

    @property
    def nick(self):
        return self._nickmask.nick

    @property
    def user(self):
        return self._nickmask.user

    @property
    def host(self):
        return self._nickmask.host

    # generic compatibility
    person = nick

    @property
    def client(self):
        return self._nickmask.userhost

    @property
    def fullname(self):
        # TODO: this should be possible to get
        return None

    @property
    def aclattr(self):
        return IRCBackend.aclpattern.format(nick=self._nickmask.nick,
                                            user=self._nickmask.user,
                                            host=self._nickmask.host)

    def __unicode__(self):
        return str(self._nickmask)

    def __str__(self):
        return self.__unicode__()

    def __eq__(self, other):
        if not isinstance(other, IRCPerson):
            log.warning("Weird you are comparing an IRCPerson to a %s.", type(other))
            return False
        return self.person == other.person


class IRCRoomOccupant(IRCPerson, RoomOccupant):

    def __init__(self, mask, room):
        super().__init__(mask)
        self._room = room

    @property
    def room(self):
        return self._room

    def __unicode__(self):
        return self._nickmask

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return f'<{self.__unicode__()} - {super().__repr__()}>'


class IRCRoom(Room):
    """
        Represent the specifics of a IRC Room/Channel.

        This lifecycle of this object is:
         - Created in IRCConnection.on_join
         - The joined status change in IRCConnection on_join/on_part
         - Deleted/destroyed in IRCConnection.on_disconnect
    """

    def __init__(self, room, bot):
        self._bot = bot
        self.room = room
        self.connection = self._bot.conn.connection
        self._topic_lock = threading.Lock()
        self._topic = None

    def __unicode__(self):
        return self.room

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return f"<{self.__unicode__()} - {super().__repr__()}>"

    def cb_set_topic(self, current_topic):
        """
        Store the current topic for this room.

        This method is called by the IRC backend when a `currenttopic`,
        `topic` or `notopic` IRC event is received to store the topic set for this channel.

        This function is not meant to be executed by regular plugins.
        To get or set
        """
        with self._topic_lock:
            self._topic = current_topic

    def join(self, username=None, password=None):
        """
        Join the room.

        If the room does not exist yet, this will automatically call
        :meth:`create` on it first.
        """
        if username is not None:
            log.debug("Ignored username parameter on join(), it is unsupported on this back-end.")
        if password is None:
            password = ""  # nosec

        self.connection.join(self.room, key=password)
        log.info('Joined room %s.', self.room)

    def leave(self, reason=None):
        """
        Leave the room.

        :param reason:
            An optional string explaining the reason for leaving the room
        """
        if reason is None:
            reason = ""

        self.connection.part(self.room, reason)
        log.info('Leaving room %s with reason %s.', self.room, reason if reason is not None else '')

    def create(self):
        """
        Not supported on this back-end. Will join the room to ensure it exists, instead.
        """
        logging.warning('IRC back-end does not support explicit creation, joining room instead to ensure it exists.')
        self.join()

    def destroy(self):
        """
        Not supported on IRC, will raise :class:`~errbot.backends.base.RoomError`.
        """
        raise RoomError('IRC back-end does not support destroying rooms.')

    @property
    def exists(self):
        """
        Boolean indicating whether this room already exists or not.

        :getter:
            Returns `True` if the room exists, `False` otherwise.
        """
        logging.warning('IRC back-end does not support determining if a room exists. '
                        'Returning the result of joined instead.')
        return self.joined

    @property
    def joined(self):
        """
        Boolean indicating whether this room has already been joined.

        :getter:
            Returns `True` if the room has been joined, `False` otherwise.
        """
        return self.room in self._bot.conn.channels.keys()

    @property
    def topic(self):
        """
        The room topic.

        :getter:
            Returns the topic (a string) if one is set, `None` if no
            topic has been set at all.
        """
        if not self.joined:
            raise RoomNotJoinedError('Must join the room to get the topic.')
        with self._topic_lock:
            return self._topic

    @topic.setter
    def topic(self, topic):
        """
        Set the room's topic.

        :param topic:
            The topic to set.
        """
        if not self.joined:
            raise RoomNotJoinedError('Must join the room to set the topic.')
        self.connection.topic(self.room, topic)

    @property
    def occupants(self):
        """
        The room's occupants.

        :getter:
            Returns a list of occupants.
            :raises:
            :class:`~MUCNotJoinedError` if the room has not yet been joined.
        """
        occupants = []
        try:
            for nick in self._bot.conn.channels[self.room].users():
                occupants.append(IRCRoomOccupant(nick, room=self.room))
        except KeyError:
            raise RoomNotJoinedError('Must be in a room in order to see occupants.')
        return occupants

    def invite(self, *args):
        """
        Invite one or more people into the room.

        :*args:
            One or more nicks to invite into the room.
        """
        for nick in args:
            self.connection.invite(nick, self.room)
            log.info('Invited %s to %s.', nick, self.room)

    def __eq__(self, other):
        if not isinstance(other, IRCRoom):
            log.warning('This is weird you are comparing an IRCRoom to a %s.', type(other))
            return False
        return self.room == other.room


class IRCConnection(SingleServerIRCBot):

    def __init__(self,
                 bot,
                 nickname,
                 server,
                 port=6667,
                 ssl=False,
                 bind_address=None,
                 ipv6=False,
                 password=None,
                 username=None,
                 nickserv_password=None,
                 private_rate=1,
                 channel_rate=1,
                 reconnect_on_kick=5,
                 reconnect_on_disconnect=5):
        self.use_ssl = ssl
        self.use_ipv6 = ipv6
        self.bind_address = bind_address
        self.bot = bot
        # manually decorate functions
        if private_rate:
            self.send_private_message = rate_limited(private_rate)(self.send_private_message)

        if channel_rate:
            self.send_public_message = rate_limited(channel_rate)(self.send_public_message)
        self._reconnect_on_kick = reconnect_on_kick
        self._pending_transfers = {}
        self._rooms_lock = threading.Lock()
        self._rooms = {}
        self._recently_joined_to = set()

        self.nickserv_password = nickserv_password
        if username is None:
            username = nickname
        self.transfers = {}
        super().__init__([(server, port, password)], nickname, username, reconnection_interval=reconnect_on_disconnect)

    def connect(self, *args, **kwargs):
        # Decode all input to UTF-8, but use a replacement character for
        # unrecognized byte sequences
        # (as described at https://pypi.python.org/pypi/irc)
        self.connection.buffer_class.errors = 'replace'

        connection_factory_kwargs = {}
        if self.use_ssl:
            import ssl
            connection_factory_kwargs['wrapper'] = ssl.wrap_socket
        if self.bind_address is not None:
            connection_factory_kwargs['bind_address'] = self.bind_address
        if self.use_ipv6:
            connection_factory_kwargs['ipv6'] = True

        connection_factory = irc.connection.Factory(**connection_factory_kwargs)
        self.connection.connect(*args, connect_factory=connection_factory, **kwargs)

    def on_welcome(self, _, e):
        log.info("IRC welcome %s", e)

        # try to identify with NickServ if there is a NickServ password in the
        # config
        if self.nickserv_password:
            msg = f'identify {self.nickserv_password}'
            self.send_private_message('NickServ', msg)

        # Must be done in a background thread, otherwise the join room
        # from the ChatRoom plugin joining channels from CHATROOM_PRESENCE
        # ends up blocking on connect.
        t = threading.Thread(target=self.bot.connect_callback)
        t.setDaemon(True)
        t.start()

    def _pubmsg(self, e, notice=False):
        msg = Message(e.arguments[0], extras={'notice': notice})
        room_name = e.target
        if room_name[0] != '#' and room_name[0] != '$':
            raise Exception(f'[{room_name}] is not a room')
        room = IRCRoom(room_name, self.bot)
        msg.frm = IRCRoomOccupant(e.source, room)
        msg.to = room
        msg.nick = msg.frm.nick  # FIXME find the real nick in the channel
        self.bot.callback_message(msg)

        possible_mentions = re.findall(IRC_NICK_REGEX, e.arguments[0])
        room_users = self.channels[room_name].users()
        mentions = filter(lambda x: x in room_users, possible_mentions)
        if mentions:
            mentions = [self.bot.build_identifier(mention) for mention in mentions]
            self.bot.callback_mention(msg, mentions)

    def _privmsg(self, e, notice=False):
        msg = Message(e.arguments[0], extras={'notice': notice})
        msg.frm = IRCPerson(e.source)
        msg.to = IRCPerson(e.target)
        self.bot.callback_message(msg)

    def on_pubmsg(self, _, e):
        self._pubmsg(e)

    def on_privmsg(self, _, e):
        self._privmsg(e)

    def on_pubnotice(self, _, e):
        self._pubmsg(e, True)

    def on_privnotice(self, _, e):
        self._privmsg(e, True)

    def on_kick(self, _, e):
        if not self._reconnect_on_kick:
            log.info("RECONNECT_ON_KICK is 0 or None, won't try to reconnect")
            return
        log.info('Got kicked out of %s... reconnect in %d seconds... ', e.target, self._reconnect_on_kick)

        def reconnect_channel(name):
            log.info('Reconnecting to %s after having beeing kicked.', name)
            self.bot.query_room(name).join()
        t = threading.Timer(self._reconnect_on_kick, reconnect_channel, [e.target, ])
        t.daemon = True
        t.start()

    def send_private_message(self, to, line):
        try:
            self.connection.privmsg(to, line)
        except ServerNotConnectedError:
            pass  # the message will be lost

    def send_public_message(self, to, line):
        try:
            self.connection.privmsg(to, line)
        except ServerNotConnectedError:
            pass  # the message will be lost

    def on_disconnect(self, connection, event):
        self._rooms = {}
        self.bot.disconnect_callback()

    def send_stream_request(self, identifier, fsource, name=None, size=None, stream_type=None):
        # Creates a new connection
        dcc = self.dcc_listen("raw")
        msg_parts = map(str, (
            'SEND',
            name,
            irc.client.ip_quad_to_numstr(dcc.localaddress),
            dcc.localport,
            size,
        ))
        msg = subprocess.list2cmdline(msg_parts)
        self.connection.ctcp("DCC", identifier.nick, msg)
        stream = Stream(identifier, fsource, name, size, stream_type)
        self.transfers[dcc] = stream

        return stream

    def on_dcc_connect(self, dcc, event):
        stream = self.transfers.get(dcc, None)
        if stream is None:
            log.error('DCC connect on a none registered connection')
            return
        log.debug('Start transfer for %s.', stream.identifier)
        stream.accept()
        self.send_chunk(stream, dcc)

    def on_dcc_disconnect(self, dcc, event):
        self.transfers.pop(dcc)

    def on_part(self, connection, event):
        """
            Handler of the part IRC Message/event.

            The part message is sent to the client as a confirmation of a
            /PART command sent by someone in the room/channel.
            If the event.source contains the bot nickname then we need to fire
            the :meth:`~errbot.backends.base.Backend.callback_room_left` event on the bot.

            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                The event.source contains the nickmask of the user that
                leave the room
                The event.target contains the channel name
        """
        leaving_nick = event.source.nick
        leaving_room = event.target
        if self.bot.bot_identifier.nick == leaving_nick:
            with self._rooms_lock:
                self.bot.callback_room_left(self._rooms[leaving_room])
            log.info('Left room {}.', leaving_room)

    def on_endofnames(self, connection, event):
        """
            Handler of the enfofnames IRC message/event.

            The endofnames message is sent to the client when the server finish
            to send the list of names of the room ocuppants.
            This usually happens when you join to the room.
            So in this case, we use this event to determine that our bot is
            finally joined to the room.

            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                the event.arguments[0] contains the channel name
        """
        # The event.arguments[0] contains the channel name.
        # We filter that to avoid a misfire of the event.
        room_name = event.arguments[0]
        with self._rooms_lock:
            if room_name in self._recently_joined_to:
                self._recently_joined_to.remove(room_name)
                self.bot.callback_room_joined(self._rooms[room_name])

    def on_join(self, connection, event):
        """
            Handler of the join IRC message/event.
            Is in response of a /JOIN client message.

            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                the event.target contains the channel name
        """
        # We can't fire the room_joined event yet,
        # because we don't have the occupants info.
        # We need to wait to endofnames message.
        room_name = event.target
        with self._rooms_lock:
            if room_name not in self._rooms:
                self._rooms[room_name] = IRCRoom(room_name, self.bot)
            self._recently_joined_to.add(room_name)

    def on_currenttopic(self, connection, event):
        """
            When you Join a room with a topic set this event fires up to
            with the topic information.
            If the room that you join don't have a topic set, nothing happens.
            Here is NOT the place to fire the :meth:`~errbot.backends.base.Backend.callback_room_topic` event for
            that case exist on_topic.

            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                The event.arguments[0] contains the room name
                The event.arguments[1] contains the topic of the room.
        """
        room_name, current_topic = event.arguments
        with self._rooms_lock:
            self._rooms[room_name].cb_set_topic(current_topic)

    def on_topic(self, connection, event):
        """
            On response to the /TOPIC command if the room have a topic.
            If the room don't have a topic the event fired is on_notopic
            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                The event.target contains the room name.
                The event.arguments[0] contains the topic name
        """
        room_name = event.target
        current_topic = event.arguments[0]
        with self._rooms_lock:
            self._rooms[room_name].cb_set_topic(current_topic)
            self.bot.callback_room_topic(self._rooms[room_name])

    def on_notopic(self, connection, event):
        """
            This event fires ip when there is no topic set on a room

            :param connection: Is an 'irc.client.ServerConnection' object

            :param event: Is an 'irc.client.Event' object
                The event.arguments[0] contains the room name
        """
        room_name = event.arguments[0]
        with self._rooms_lock:
            self._rooms[room_name].cb_set_topic(None)
            self.bot.callback_room_topic(self._rooms[room_name])

    @staticmethod
    def send_chunk(stream, dcc):
        data = stream.read(4096)
        dcc.send_bytes(data)
        stream.ack_data(len(data))

    def on_dccmsg(self, dcc, event):
        stream = self.transfers.get(dcc, None)
        if stream is None:
            log.error("DCC connect on a none registered connection")
            return
        acked = struct.unpack("!I", event.arguments[0])[0]
        if acked == stream.size:
            log.info('File %s successfully transfered to %s', stream.name, stream.identifier)
            dcc.disconnect()
            self.transfers.pop(dcc)
        elif acked == stream.transfered:
            log.debug('Chunk for file %s successfully transfered to %s (%d/%d).',
                      stream.name, stream.identifier, stream.transfered, stream.size)
            self.send_chunk(stream, dcc)
        else:
            log.debug('Partial chunk for file %s successfully transfered to %s (%d/%d), wait for more',
                      stream.name, stream.identifier, stream.transfered, stream.size)

    def away(self, message=''):
        """
        Extend the original implementation to support AWAY.
        To set an away message, set message to something.
        To cancel an away message, leave message at empty string.
        """
        self.connection.send_raw(' '.join(['AWAY', message]).strip())


class IRCBackend(ErrBot):
    aclpattern = '{nick}!{user}@{host}'

    def __init__(self, config):
        if hasattr(config, 'IRC_ACL_PATTERN'):
            IRCBackend.aclpattern = config.IRC_ACL_PATTERN

        identity = config.BOT_IDENTITY
        nickname = identity['nickname']
        server = identity['server']
        port = identity.get('port', 6667)
        password = identity.get('password', None)
        ssl = identity.get('ssl', False)
        bind_address = identity.get('bind_address', None)
        ipv6 = identity.get('ipv6', False)
        username = identity.get('username', None)
        nickserv_password = identity.get('nickserv_password', None)

        compact = config.COMPACT_OUTPUT if hasattr(config, 'COMPACT_OUTPUT') else True
        enable_format('irc', IRC_CHRS, borders=not compact)

        private_rate = getattr(config, 'IRC_PRIVATE_RATE', 1)
        channel_rate = getattr(config, 'IRC_CHANNEL_RATE', 1)
        reconnect_on_kick = getattr(config, 'IRC_RECONNECT_ON_KICK', 5)
        reconnect_on_disconnect = getattr(config, 'IRC_RECONNECT_ON_DISCONNECT', 5)

        self.bot_identifier = IRCPerson(nickname + '!' + nickname + '@' + server)
        super().__init__(config)
        self.conn = IRCConnection(bot=self,
                                  nickname=nickname,
                                  server=server,
                                  port=port,
                                  ssl=ssl,
                                  bind_address=bind_address,
                                  ipv6=ipv6,
                                  password=password,
                                  username=username,
                                  nickserv_password=nickserv_password,
                                  private_rate=private_rate,
                                  channel_rate=channel_rate,
                                  reconnect_on_kick=reconnect_on_kick,
                                  reconnect_on_disconnect=reconnect_on_disconnect,
                                  )
        self.md = irc_md()
        config.MESSAGE_SIZE_LIMIT = IRC_MESSAGE_SIZE_LIMIT

    def send_message(self, msg):
        super().send_message(msg)
        if msg.is_direct:
            msg_func = self.conn.send_private_message
            msg_to = msg.to.person
        else:
            msg_func = self.conn.send_public_message
            msg_to = msg.to.room

        body = self.md.convert(msg.body)
        for line in body.split('\n'):
            msg_func(msg_to, line)

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        if status == ONLINE:
            self.conn.away()  # cancels the away message
        else:
            self.conn.away(f'[{status}] {message}')

    def send_stream_request(self, identifier, fsource, name=None, size=None, stream_type=None):
        return self.conn.send_stream_request(identifier, fsource, name, size, stream_type)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = self.build_message(text)
        if msg.is_group:
            if private:
                response.frm = self.bot_identifier
                response.to = IRCPerson(str(msg.frm))
            else:
                response.frm = IRCRoomOccupant(str(self.bot_identifier), msg.frm.room)
                response.to = msg.frm.room
        else:
            response.frm = self.bot_identifier
            response.to = msg.frm
        return response

    def serve_forever(self):
        try:
            self.conn.start()
        except KeyboardInterrupt:
            log.info("Interrupt received, shutting down")
        finally:
            self.conn.disconnect("Shutting down")
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    def connect(self):
        return self.conn

    def build_message(self, text):
        text = text.replace('', '*')  # there is a weird chr IRC is sending that we need to filter out
        return super().build_message(text)

    def build_identifier(self, txtrep):
        log.debug('Build identifier from %s.', txtrep)
        # A textual representation starting with # means that we are talking
        # about an IRC channel -- IRCRoom in internal err-speak.
        if txtrep.startswith('#'):
            return IRCRoom(txtrep, self)

        # Occupants are represented as 2 lines, one is the IRC mask and the second is the Room.
        if '\n' in txtrep:
            m, r = txtrep.split('\n')
            return IRCRoomOccupant(m, IRCRoom(r, self))
        return IRCPerson(txtrep)

    def shutdown(self):
        super().shutdown()

    def query_room(self, room):
        """
        Query a room for information.

        :param room:
            The channel name to query for.
        :returns:
            An instance of :class:`~IRCMUCRoom`.
        """
        with self.conn._rooms_lock:
            if room not in self.conn._rooms:
                self.conn._rooms[room] = IRCRoom(room, self)
            return self.conn._rooms[room]

    @property
    def mode(self):
        return 'irc'

    def rooms(self):
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~IRCMUCRoom` instances.
        """
        with self.conn._rooms_lock:
            return self.conn._rooms.values()

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'{identifier.nick}: {message.body}'



from errbot import BotPlugin, botcmd


class WhateverName(BotPlugin):
    """Test plugin to verify that now class names don't matter.
    """
    @botcmd
    def myname(self, msg, args):
        return self.name



import logging

import pytest
from errbot.core import ErrBot
from errbot.bootstrap import CORE_BACKENDS
from errbot.backend_plugin_manager import BackendPluginManager

logging.basicConfig(level=logging.DEBUG)

backends_to_check = ['Text', 'Test', 'Null']


@pytest.mark.parametrize('backend_name', backends_to_check)
def test_builtins(backend_name):
    bpm = BackendPluginManager({}, 'errbot.backends', backend_name, ErrBot, CORE_BACKENDS)
    assert bpm.plugin_info.name == backend_name



import types
from collections.abc import MutableMapping
from contextlib import contextmanager
import logging
log = logging.getLogger(__name__)


class StoreException(Exception):
    pass


class StoreAlreadyOpenError(StoreException):
    pass


class StoreNotOpenError(StoreException):
    pass


class StoreMixin(MutableMapping):
    """
     This class handle the basic needs of bot plugins and core like loading, unloading and creating a storage
    """

    def __init__(self):
        self._store = None
        self.namespace = None

    def open_storage(self, storage_plugin, namespace):
        if hasattr(self, 'store') and self._store is not None:
            raise StoreAlreadyOpenError("Storage appears to be opened already")
        log.debug("Opening storage '%s'", namespace)
        self._store = storage_plugin.open(namespace)
        self.namespace = namespace

    def close_storage(self):
        if not hasattr(self, '_store') or self._store is None:
            raise StoreNotOpenError("Storage does not appear to have been opened yet")
        self._store.close()
        self._store = None
        log.debug("Closed storage '%s'", self.namespace)

    # those are the minimal things to behave like a dictionary with the UserDict.DictMixin
    def __getitem__(self, key):
        return self._store.get(key)

    @contextmanager
    def mutable(self, key, default=None):
        try:
            obj = self._store.get(key)
        except KeyError:
            obj = default
        yield obj
        # implements autosave for a plugin persistent entry
        # with self['foo'] as f:
        #     f[4] = 2
        # saves the entry !
        self._store.set(key, obj)

    def __setitem__(self, key, item):
        return self._store.set(key, item)

    def __delitem__(self, key):
        return self._store.remove(key)

    def keys(self):
        return self._store.keys()

    def __len__(self):
        return self._store.len()

    def __iter__(self):
        for i in self._store.keys():
            yield i

    def __contains__(self, x):
        try:
            self._store.get(x)
            return True
        except KeyError:
            return False

    # compatibility with with
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close_storage()



import os

extra_plugin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'room_plugin')
extra_config = {'CORE_PLUGINS': ('Help', 'Utils', 'CommandNotFoundFilter'),
                'BOT_ALT_PREFIXES': ('!',),
                'BOT_PREFIX': '$'}


def test_help_is_still_here(testbot):
    assert 'All commands' in testbot.exec_command('!help')


def test_backup_help_not_here(testbot):
    assert 'That command is not defined.' in testbot.exec_command('!help backup')


def test_backup_should_not_be_there(testbot):
    assert 'Command "backup" not found.' in testbot.exec_command('!backup')


def test_echo_still_here(testbot):
    assert 'toto' in testbot.exec_command('!echo toto')


def test_bot_prefix_replaced(testbot):
    assert '$help - Returns a help string' in testbot.exec_command('$help')



from __future__ import absolute_import
from errbot import BotPlugin, botcmd, re_botcmd, arg_botcmd


class Test(BotPlugin):
    @botcmd
    def test_template1(self, msg, args):
        self.send_templated(msg.frm, 'test', {'variable': 'ok'})

    @botcmd(template='test')
    def test_template2(self, msg, args):
        return {'variable': 'ok'}

    @botcmd(template='test')
    def test_template3(self, msg, args):
        yield {'variable': 'ok'}

    @arg_botcmd('my_var', type=str, template='test')
    def test_template4(self, msg, my_var=None):
        return {'variable': my_var}



# coding=utf-8
import sys
import logging
from pathlib import Path
from tempfile import mkdtemp
from os.path import sep

import pytest
import os  # noqa
import re  # noqa
from collections import OrderedDict
from queue import Queue, Empty  # noqa
from errbot.core import ErrBot
from errbot.backends.base import Message, Room, Identifier, ONLINE
from errbot.backends.test import TestPerson, TestOccupant, TestRoom, ShallowConfig
from errbot import botcmd, re_botcmd, arg_botcmd, templating  # noqa
from errbot.bootstrap import CORE_STORAGE, bot_config_defaults
from errbot.plugin_manager import BotPluginManager
from errbot.rendering import text
from errbot.core_plugins.acls import ACLS
from errbot.repo_manager import BotRepoManager
from errbot.backend_plugin_manager import BackendPluginManager
from errbot.storage.base import StoragePluginBase
from errbot.utils import PLUGINS_SUBDIR

LONG_TEXT_STRING = "This is a relatively long line of output, but I am repeated multiple times.\n"

logging.basicConfig(level=logging.DEBUG)

SIMPLE_JSON_PLUGINS_INDEX = """
{"errbotio/err-helloworld":
    {"HelloWorld":
        {"path": "/helloWorld.plug",
         "documentation": "let's say hello !",
         "avatar_url": "https://avatars.githubusercontent.com/u/15802630?v=3",
         "name": "HelloWorld",
         "python": "2+",
         "repo": "https://github.com/errbotio/err-helloworld"
         }
    }
}
"""


class DummyBackend(ErrBot):

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        pass

    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
        pass

    def query_room(self, room: str) -> Room:
        pass

    def __init__(self, extra_config=None):
        self.outgoing_message_queue = Queue()
        if extra_config is None:
            extra_config = {}
        # make up a config.
        tempdir = mkdtemp()
        # reset the config every time
        sys.modules.pop('errbot.config-template', None)
        __import__('errbot.config-template')
        config = ShallowConfig()
        config.__dict__.update(sys.modules['errbot.config-template'].__dict__)
        bot_config_defaults(config)

        # It injects itself as a plugin. Changed the name to be sure we distinguish it.
        self.name = 'DummyBackendRealName'

        config.BOT_DATA_DIR = tempdir
        config.BOT_LOG_FILE = tempdir + sep + 'log.txt'
        config.BOT_PLUGIN_INDEXES = tempdir + sep + 'repos.json'
        config.BOT_EXTRA_PLUGIN_DIR = []
        config.BOT_LOG_LEVEL = logging.DEBUG
        config.BOT_IDENTITY = {'username': 'err@localhost'}
        config.BOT_ASYNC = False
        config.BOT_PREFIX = '!'
        config.CHATROOM_FN = 'blah'

        # Writeout the made up repos file
        with open(config.BOT_PLUGIN_INDEXES, "w") as index_file:
            index_file.write(SIMPLE_JSON_PLUGINS_INDEX)

        for key in extra_config:
            setattr(config, key, extra_config[key])
        super().__init__(config)
        self.bot_identifier = self.build_identifier('err')
        self.md = text()  # We just want simple text for testing purposes

        # setup a memory based storage
        spm = BackendPluginManager(config, 'errbot.storage', 'Memory', StoragePluginBase, CORE_STORAGE)
        storage_plugin = spm.load_plugin()

        # setup the plugin_manager just internally
        botplugins_dir = os.path.join(config.BOT_DATA_DIR, PLUGINS_SUBDIR)
        if not os.path.exists(botplugins_dir):
            os.makedirs(botplugins_dir, mode=0o755)

        # get it back from where we publish it.
        repo_index_paths = (os.path.join(os.path.dirname(__file__), '..', 'docs', '_extra', 'repos.json'),)
        repo_manager = BotRepoManager(storage_plugin,
                                      botplugins_dir,
                                      repo_index_paths)
        self.attach_storage_plugin(storage_plugin)
        self.attach_repo_manager(repo_manager)
        self.attach_plugin_manager(BotPluginManager(storage_plugin,
                                                    config.BOT_EXTRA_PLUGIN_DIR,
                                                    config.AUTOINSTALL_DEPS,
                                                    getattr(config, 'CORE_PLUGINS', None),
                                                    lambda name, clazz: clazz(self, name),
                                                    getattr(config, 'PLUGINS_CALLBACK_ORDER', (None, ))))
        self.inject_commands_from(self)
        self.inject_command_filters_from(ACLS(self))

    def build_identifier(self, text_representation):
        return TestPerson(text_representation)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        reply = self.build_message(text)
        reply.frm = self.bot_identifier
        reply.to = msg.frm
        if threaded:
            reply.parent = msg
        return reply

    def send_message(self, msg):
        msg._body = self.md.convert(msg.body)
        self.outgoing_message_queue.put(msg)

    def pop_message(self, timeout=3, block=True):
        return self.outgoing_message_queue.get(timeout=timeout, block=block)

    @botcmd
    def command(self, msg, args):
        return "Regular command"

    @botcmd(admin_only=True)
    def admin_command(self, msg, args):
        return "Admin command"

    @re_botcmd(pattern=r'^regex command with prefix$', prefixed=True)
    def regex_command_with_prefix(self, msg, match):
        return "Regex command"

    @re_botcmd(pattern=r'^regex command without prefix$', prefixed=False)
    def regex_command_without_prefix(self, msg, match):
        return "Regex command"

    @re_botcmd(pattern=r'regex command with capture group: (?P<capture>.*)', prefixed=False)
    def regex_command_with_capture_group(self, msg, match):
        return match.group('capture')

    @re_botcmd(pattern=r'matched by two commands')
    def double_regex_command_one(self, msg, match):
        return "one"

    @re_botcmd(pattern=r'matched by two commands', flags=re.IGNORECASE)
    def double_regex_command_two(self, msg, match):
        return "two"

    @re_botcmd(pattern=r'match_here', matchall=True)
    def regex_command_with_matchall(self, msg, matches):
        return len(matches)

    @botcmd
    def return_args_as_str(self, msg, args):
        return "".join(args)

    @botcmd(template='args_as_md')
    def return_args_as_md(self, msg, args):
        return {'args': args}

    @botcmd
    def send_args_as_md(self, msg, args):
        self.send_templated(msg.frm, 'args_as_md', {'args': args})

    @botcmd
    def raises_exception(self, msg, args):
        raise Exception("Kaboom!")

    @botcmd
    def yield_args_as_str(self, msg, args):
        for arg in args:
            yield arg

    @botcmd(template='args_as_md')
    def yield_args_as_md(self, msg, args):
        for arg in args:
            yield {'args': [arg]}

    @botcmd
    def yields_str_then_raises_exception(self, msg, args):
        yield 'foobar'
        raise Exception('Kaboom!')

    @botcmd
    def return_long_output(self, msg, args):
        return LONG_TEXT_STRING * 3

    @botcmd
    def yield_long_output(self, msg, args):
        for i in range(2):
            yield LONG_TEXT_STRING * 3

    ##
    # arg_botcmd test commands
    ##

    @arg_botcmd('--first-name', dest='first_name')
    @arg_botcmd('--last-name', dest='last_name')
    def yields_first_name_last_name(self, msg, first_name=None, last_name=None):
        yield "%s %s" % (first_name, last_name)

    @arg_botcmd('--first-name', dest='first_name')
    @arg_botcmd('--last-name', dest='last_name')
    def returns_first_name_last_name(self, msg, first_name=None, last_name=None):
        return "%s %s" % (first_name, last_name)

    @arg_botcmd('--first-name', dest='first_name')
    @arg_botcmd('--last-name', dest='last_name', unpack_args=False)
    def returns_first_name_last_name_without_unpacking(self, msg, args):
        return "%s %s" % (args.first_name, args.last_name)

    @arg_botcmd('value', type=str)
    @arg_botcmd('--count', dest='count', type=int)
    def returns_value_repeated_count_times(self, msg, value=None, count=None):
        # str * int gives a repeated string
        return value * count

    @property
    def mode(self):
        return "Dummy"

    @property
    def rooms(self):
        return []


@pytest.fixture
def dummy_backend():
    return DummyBackend()


def test_buildreply(dummy_backend):
    m = dummy_backend.build_message('Content')
    m.frm = dummy_backend.build_identifier('user')
    m.to = dummy_backend.build_identifier('somewhere')
    resp = dummy_backend.build_reply(m, 'Response')

    assert str(resp.to) == 'user'
    assert str(resp.frm) == 'err'
    assert str(resp.body) == 'Response'
    assert resp.parent is None


def test_buildreply_with_parent(dummy_backend):
    m = dummy_backend.build_message('Content')
    m.frm = dummy_backend.build_identifier('user')
    m.to = dummy_backend.build_identifier('somewhere')
    resp = dummy_backend.build_reply(m, 'Response', threaded=True)

    assert resp.parent is not None


def test_bot_admins_unique_string():
    dummy = DummyBackend(extra_config={'BOT_ADMINS': 'err@localhost'})
    assert dummy.bot_config.BOT_ADMINS == ('err@localhost',)


@pytest.fixture
def dummy_execute_and_send():
    dummy = DummyBackend()
    example_message = dummy.build_message('some_message')
    example_message.frm = dummy.build_identifier('noterr')
    example_message.to = dummy.build_identifier('err')

    assets_path = os.path.join(os.path.dirname(__file__), 'assets')
    templating.template_path.append(str(templating.make_templates_path(Path(assets_path))))
    templating.env = templating.Environment(loader=templating.FileSystemLoader(templating.template_path))
    return dummy, example_message


def test_commands_can_return_string(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send

    dummy._execute_and_send(cmd='return_args_as_str', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.return_args_as_str._err_command_template)
    assert "foobar" == dummy.pop_message().body


def test_commands_can_return_md(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy._execute_and_send(cmd='return_args_as_md', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.return_args_as_md._err_command_template)
    response = dummy.pop_message()
    assert "foobar" == response.body


def test_commands_can_send_templated(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy._execute_and_send(cmd='send_args_as_md', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.return_args_as_md._err_command_template)
    response = dummy.pop_message()
    assert "foobar" == response.body


def test_exception_is_caught_and_shows_error_message(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy._execute_and_send(cmd='raises_exception', args=[], match=None, msg=m,
                            template_name=dummy.raises_exception._err_command_template)
    assert dummy.MSG_ERROR_OCCURRED in dummy.pop_message().body

    dummy._execute_and_send(cmd='yields_str_then_raises_exception', args=[], match=None, msg=m,
                            template_name=dummy.yields_str_then_raises_exception._err_command_template)
    assert "foobar" == dummy.pop_message().body
    assert dummy.MSG_ERROR_OCCURRED in dummy.pop_message().body


def test_commands_can_yield_strings(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy._execute_and_send(cmd='yield_args_as_str', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.yield_args_as_str._err_command_template)
    assert "foo" == dummy.pop_message().body
    assert "bar" == dummy.pop_message().body


def test_commands_can_yield_md(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy._execute_and_send(cmd='yield_args_as_md', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.yield_args_as_md._err_command_template)
    assert "foo" == dummy.pop_message().body
    assert "bar" == dummy.pop_message().body


def test_output_longer_than_max_msg_size_is_split_into_multiple_msgs_when_returned(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy.bot_config.MESSAGE_SIZE_LIMIT = len(LONG_TEXT_STRING)

    dummy._execute_and_send(cmd='return_long_output', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.return_long_output._err_command_template)
    for i in range(3):  # return_long_output outputs a string that's 3x longer than the size limit
        assert LONG_TEXT_STRING.strip() == dummy.pop_message().body

    with pytest.raises(Empty):
        dummy.pop_message(block=False)


def test_output_longer_than_max_msg_size_is_split_into_multiple_msgs_when_yielded(dummy_execute_and_send):
    dummy, m = dummy_execute_and_send
    dummy.bot_config.MESSAGE_SIZE_LIMIT = len(LONG_TEXT_STRING)

    dummy._execute_and_send(cmd='yield_long_output', args=['foo', 'bar'], match=None, msg=m,
                            template_name=dummy.yield_long_output._err_command_template)
    for i in range(6):  # yields_long_output yields 2 strings that are 3x longer than the size limit
        assert LONG_TEXT_STRING.strip() == dummy.pop_message().body
    with pytest.raises(Empty):
        dummy.pop_message(block=False)


def makemessage(dummy, message, from_=None, to=None):
    if not from_:
        from_ = dummy.build_identifier("noterr")
    if not to:
        to = dummy.build_identifier("noterr")
    m = dummy.build_message(message)
    m.frm = from_
    m.to = to
    return m


def test_inject_skips_methods_without_botcmd_decorator(dummy_backend):
    assert 'build_message' not in dummy_backend.commands


def test_inject_and_remove_botcmd(dummy_backend):
    assert 'command' in dummy_backend.commands
    dummy_backend.remove_commands_from(dummy_backend)
    assert len(dummy_backend.commands) == 0


def test_inject_and_remove_re_botcmd(dummy_backend):
    assert 'regex_command_with_prefix' in dummy_backend.re_commands
    dummy_backend.remove_commands_from(dummy_backend)
    assert len(dummy_backend.re_commands) == 0


def test_callback_message(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!return_args_as_str one two"))
    assert "one two" == dummy_backend.pop_message().body


def test_callback_message_with_prefix_optional():
    dummy = DummyBackend({'BOT_PREFIX_OPTIONAL_ON_CHAT': True})
    m = makemessage(dummy, "return_args_as_str one two")
    dummy.callback_message(m)
    assert "one two" == dummy.pop_message().body

    # Groupchat should still require the prefix
    m.frm = TestOccupant("someone", "room")
    room = TestRoom("room", bot=dummy)
    m.to = room
    dummy.callback_message(m)

    with pytest.raises(Empty):
        dummy.pop_message(block=False)

    m = makemessage(dummy, "!return_args_as_str one two",
                    from_=TestOccupant("someone", "room"),
                    to=room)
    dummy.callback_message(m)
    assert "one two" == dummy.pop_message().body


def test_callback_message_with_bot_alt_prefixes():
    dummy = DummyBackend({'BOT_ALT_PREFIXES': ('Err',), 'BOT_ALT_PREFIX_SEPARATORS': (',', ';')})
    dummy.callback_message(makemessage(dummy, "Err return_args_as_str one two"))
    assert "one two" == dummy.pop_message().body
    dummy.callback_message(makemessage(dummy, "Err, return_args_as_str one two"))
    assert "one two" == dummy.pop_message().body


def test_callback_message_with_re_botcmd(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!regex command with prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "regex command without prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "!regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "This command also allows extra text in front - regex "
                                                              "command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body


def test_callback_message_with_re_botcmd_and_alt_prefixes():
    dummy_backend = DummyBackend({'BOT_ALT_PREFIXES': ('Err',), 'BOT_ALT_PREFIX_SEPARATORS': (',', ';')})
    dummy_backend.callback_message(makemessage(dummy_backend, "!regex command with prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "Err regex command with prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "Err, regex command with prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "regex command without prefix"))
    assert "Regex command" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "!regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "This command also allows extra text in front - "
                                                              "regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "Err, regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "Err This command also allows extra text in front - "
                                                              "regex command with capture group: Captured text"))
    assert "Captured text" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "!match_here"))
    assert "1" == dummy_backend.pop_message().body
    dummy_backend.callback_message(makemessage(dummy_backend, "!match_here match_here match_here"))
    assert "3" == dummy_backend.pop_message().body


def test_regex_commands_can_overlap(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!matched by two commands"))
    response = (dummy_backend.pop_message().body, dummy_backend.pop_message().body)
    assert response == ("one", "two") or response == ("two", "one")


def test_regex_commands_allow_passing_re_flags(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!MaTcHeD By TwO cOmMaNdS"))
    assert "two" == dummy_backend.pop_message().body
    with pytest.raises(Empty):
        dummy_backend.pop_message(timeout=1)


def test_arg_botcmd_returns_first_name_last_name(dummy_backend):
    dummy_backend.callback_message(
        makemessage(
            dummy_backend,
            "!returns_first_name_last_name --first-name=Err --last-name=Bot"
        )
    )
    assert "Err Bot"


def test_arg_botcmd_returns_with_escaping(dummy_backend):
    first_name = 'Err\\"'
    last_name = 'Bot'
    dummy_backend.callback_message(
        makemessage(
            dummy_backend,
            "!returns_first_name_last_name --first-name=%s --last-name=%s" % (first_name, last_name)
        )
    )
    assert 'Err" Bot' == dummy_backend.pop_message().body


def test_arg_botcmd_returns_with_incorrect_escaping(dummy_backend):
    first_name = 'Err"'
    last_name = 'Bot'
    dummy_backend.callback_message(
        makemessage(
            dummy_backend,
            "!returns_first_name_last_name --first-name=%s --last-name=%s" % (first_name, last_name)
        )
    )
    assert 'I couldn\'t parse this command; No closing quotation' in dummy_backend.pop_message().body


def test_arg_botcmd_yields_first_name_last_name(dummy_backend):
    dummy_backend.callback_message(
        makemessage(
            dummy_backend,
            "!yields_first_name_last_name --first-name=Err --last-name=Bot"
        )
    )
    assert "Err Bot" == dummy_backend.pop_message().body


def test_arg_botcmd_returns_value_repeated_count_times(dummy_backend):
    dummy_backend.callback_message(
        makemessage(dummy_backend, "!returns_value_repeated_count_times Foo --count 5")
    )
    assert "FooFooFooFooFoo" == dummy_backend.pop_message().body


def test_arg_botcmd_doesnt_raise_systemerror(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!returns_first_name_last_name --invalid-parameter"))


def test_arg_botcdm_returns_errors_as_chat(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!returns_first_name_last_name --invalid-parameter"))
    assert "I couldn't parse the arguments; unrecognized arguments: --invalid-parameter" \
        in dummy_backend.pop_message().body


def test_arg_botcmd_returns_help_message_as_chat(dummy_backend):
    dummy_backend.callback_message(makemessage(dummy_backend, "!returns_first_name_last_name --help"))
    assert "usage: returns_first_name_last_name [-h] [--last-name LAST_NAME]" in dummy_backend.pop_message().body


def test_arg_botcmd_undoes_fancy_unicode_dash_conversion(dummy_backend):
    dummy_backend.callback_message(
        makemessage(
            dummy_backend,
            "!returns_first_name_last_name —first-name=Err —last-name=Bot"
        )
    )
    assert "Err Bot" == dummy_backend.pop_message().body


def test_arg_botcmd_without_argument_unpacking(dummy_backend):
    dummy_backend.callback_message(
        makemessage(dummy_backend, "!returns_first_name_last_name_without_unpacking --first-name=Err --last-name=Bot")
    )
    assert "Err Bot" == dummy_backend.pop_message().body


def test_access_controls(dummy_backend):
    testroom = TestRoom("room", bot=dummy_backend)
    tests = [
        # BOT_ADMINS scenarios
        dict(
            message=makemessage(dummy_backend, "!admin_command"),
            bot_admins=('noterr',),
            expected_response="Admin command",
        ),
        dict(
            message=makemessage(dummy_backend, "!admin_command"),
            bot_admins=(),
            expected_response="This command requires bot-admin privileges",
        ),
        dict(
            message=makemessage(dummy_backend, "!admin_command"),
            bot_admins=('*err',),
            expected_response="Admin command",
        ),

        # admin_only commands SHOULD be private-message only by default
        dict(
            message=makemessage(dummy_backend, "!admin_command",
                                from_=TestOccupant('noterr', room=testroom),
                                to=testroom),
            bot_admins=('noterr',),
            expected_response="This command may only be issued through a direct message",
        ),
        # But MAY be sent via groupchat IF 'allowmuc' is specifically set to True.
        dict(
            message=makemessage(dummy_backend, "!admin_command",
                                from_=TestOccupant('noterr', room=testroom),
                                to=testroom),
            bot_admins=('noterr',),
            acl={'admin_command': {'allowmuc': True}},
            expected_response="Admin command",
        ),

        # ACCESS_CONTROLS scenarios WITHOUT wildcards (<4.0 format)
        dict(
            message=makemessage(dummy_backend, "!command"),
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!regex command with prefix"),
            expected_response="Regex command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl_default={'allowmuc': False, 'allowprivate': False},
            expected_response="You're not allowed to access this command via private message to me"
        ),
        dict(
            message=makemessage(dummy_backend, "regex command without prefix"),
            acl_default={'allowmuc': False, 'allowprivate': False},
            expected_response="You're not allowed to access this command via private message to me"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl_default={'allowmuc': True, 'allowprivate': False},
            expected_response="You're not allowed to access this command via private message to me"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl_default={'allowmuc': False, 'allowprivate': True},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowprivate': False}},
            acl_default={'allowmuc': False, 'allowprivate': True},
            expected_response="You're not allowed to access this command via private message to me"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowmuc': True}},
            acl_default={'allowmuc': True, 'allowprivate': False},
            expected_response="You're not allowed to access this command via private message to me"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowprivate': True}},
            acl_default={'allowmuc': False, 'allowprivate': False},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room"),
                                to=TestRoom("room", bot=dummy_backend)),
            acl={'command': {'allowrooms': ('room',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room_1"),
                                to=TestRoom("room1", bot=dummy_backend)),
            acl={'command': {'allowrooms': ('room_*',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room"),
                                to=TestRoom("room", bot=dummy_backend)),
            acl={'command': {'allowrooms': ('anotherroom@localhost',)}},
            expected_response="You're not allowed to access this command from this room",
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room"),
                                to=TestRoom("room", bot=dummy_backend)),
            acl={'command': {'denyrooms': ('room',)}},
            expected_response="You're not allowed to access this command from this room",
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room"),
                                to=TestRoom("room", bot=dummy_backend)),
            acl={'command': {'denyrooms': ('*',)}},
            expected_response="You're not allowed to access this command from this room",
        ),
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=TestOccupant("someone", "room"),
                                to=TestRoom("room", bot=dummy_backend)),
            acl={'command': {'denyrooms': ('anotherroom',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowusers': ('noterr',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowusers': 'noterr'}},  # simple string instead of tuple
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowusers': ('err',)}},
            expected_response="You're not allowed to access this command from this user",
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'allowusers': ('*err',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'denyusers': ('err',)}},
            expected_response="Regular command"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'denyusers': ('noterr',)}},
            expected_response="You're not allowed to access this command from this user"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'denyusers': 'noterr'}},   # simple string instead of tuple
            expected_response="You're not allowed to access this command from this user"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'command': {'denyusers': ('*err',)}},
            expected_response="You're not allowed to access this command from this user"
        ),

        # ACCESS_CONTROLS scenarios WITH wildcards (>=4.0 format)
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'DummyBackendRealName:command': {'denyusers': ('noterr',)}},
            expected_response="You're not allowed to access this command from this user"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'*:command': {'denyusers': ('noterr',)}},
            expected_response="You're not allowed to access this command from this user"
        ),
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl={'DummyBackendRealName:*': {'denyusers': ('noterr',)}},
            expected_response="You're not allowed to access this command from this user"
        ),
        # Overlapping globs should use first match
        dict(
            message=makemessage(dummy_backend, "!command"),
            acl=OrderedDict([
                ('DummyBackendRealName:*', {'denyusers': ('noterr',)}),
                ('DummyBackendRealName:command', {'denyusers': ()})
            ]),
            expected_response="You're not allowed to access this command from this user"
        ),

        # ACCESS_CONTROLS with numeric username as in telegram
        dict(
            message=makemessage(dummy_backend, "!command",
                                from_=dummy_backend.build_identifier(1234)),
            acl={'command': {'allowusers': (1234, )}},
            expected_response="Regular command"
        ),
    ]

    for test in tests:
        dummy_backend.bot_config.ACCESS_CONTROLS_DEFAULT = test.get('acl_default', {})
        dummy_backend.bot_config.ACCESS_CONTROLS = test.get('acl', {})
        dummy_backend.bot_config.BOT_ADMINS = test.get('bot_admins', ())
        logger = logging.getLogger(__name__)
        logger.info("** message: {}".format(test['message'].body))
        logger.info("** bot_admins: {}".format(dummy_backend.bot_config.BOT_ADMINS))
        logger.info("** acl: {!r}".format(dummy_backend.bot_config.ACCESS_CONTROLS))
        logger.info("** acl_default: {!r}".format(dummy_backend.bot_config.ACCESS_CONTROLS_DEFAULT))
        dummy_backend.callback_message(test['message'])
        assert test['expected_response'] == dummy_backend.pop_message().body



from io import BytesIO
from errbot.backends.test import TestPerson
from errbot.streaming import Tee
from errbot.backends.base import Stream


class StreamingClient(object):
    def callback_stream(self, stream):
        self.response = stream.read()


def test_streaming():
    canary = b'this is my test' * 1000
    source = Stream(TestPerson("gbin@gootz.net"), BytesIO(canary))
    clients = [StreamingClient() for _ in range(50)]
    Tee(source, clients).run()
    for client in clients:
        assert client.response == canary



from os import path
from errbot.botplugin import recurse_check_structure, ValidationException
import pytest

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'config_plugin')


def test_recurse_check_structure_valid():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list=["Foo", "Bar", "Bas"], dict={'foo': "Bar"}, none=None, true=True,
                    false=False)
    recurse_check_structure(sample, to_check)


def test_recurse_check_structure_missingitem():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_recurse_check_structure_extrasubitem():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list=["Foo", "Bar", "Bas"], dict={'foo': "Bar", 'Bar': "Foo"}, none=None,
                    true=True, false=False)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_recurse_check_structure_missingsubitem():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list=["Foo", "Bar", "Bas"], dict={}, none=None, true=True, false=False)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_recurse_check_structure_wrongtype_1():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string=None, list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_recurse_check_structure_wrongtype_2():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list={'foo': "Bar"}, dict={'foo': "Bar"}, none=None, true=True, false=False)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_recurse_check_structure_wrongtype_3():
    sample = dict(string="Foobar", list=["Foo", "Bar"], dict={'foo': "Bar"}, none=None, true=True, false=False)
    to_check = dict(string="Foobar", list=["Foo", "Bar"], dict=["Foo", "Bar"], none=None, true=True, false=False)
    with pytest.raises(ValidationException):
        recurse_check_structure(sample, to_check)


def test_failed_config(testbot):
    assert 'Plugin configuration done.' in testbot.exec_command('!plugin config Config {"One": "two"}')



import gc
import os
import signal

from datetime import datetime
from errbot import BotPlugin, botcmd, arg_botcmd
from errbot.utils import format_timedelta, global_restart


class Health(BotPlugin):
    @botcmd(template='status')
    def status(self, msg, args):
        """ If I am alive I should be able to respond to this one
        """
        plugins_statuses = self.status_plugins(msg, args)
        loads = self.status_load(msg, args)
        gc = self.status_gc(msg, args)

        return {'plugins_statuses': plugins_statuses['plugins_statuses'],
                'loads': loads['loads'],
                'gc': gc['gc']}

    @botcmd(template='status_load')
    def status_load(self, _, args):
        """ shows the load status
        """
        try:
            from posix import getloadavg
            loads = getloadavg()
        except Exception:
            loads = None

        return {'loads': loads}

    @botcmd(template='status_gc')
    def status_gc(self, _, args):
        """ shows the garbage collection details
        """
        return {'gc': gc.get_count()}

    @botcmd(template='status_plugins')
    def status_plugins(self, _, args):
        """ shows the plugin status
        """
        pm = self._bot.plugin_manager
        all_blacklisted = pm.get_blacklisted_plugin()
        all_loaded = pm.get_all_active_plugin_names()
        all_attempted = sorted(pm.plugin_infos.keys())
        plugins_statuses = []
        for name in all_attempted:
            if name in all_blacklisted:
                if name in all_loaded:
                    plugins_statuses.append(('BA', name))
                else:
                    plugins_statuses.append(('BD', name))
            elif name in all_loaded:
                plugins_statuses.append(('A', name))
            elif pm.get_plugin_obj_by_name(name) is not None \
                    and pm.get_plugin_obj_by_name(name).get_configuration_template() is not None \
                    and pm.get_plugin_configuration(name) is None:
                plugins_statuses.append(('C', name))
            else:
                plugins_statuses.append(('D', name))

        return {'plugins_statuses': plugins_statuses}

    @botcmd
    def uptime(self, _, args):
        """ Return the uptime of the bot
        """
        u = format_timedelta(datetime.now() - self._bot.startup_time)
        since = self._bot.startup_time.strftime('%A, %b %d at %H:%M')
        return f"I've been up for {u} (since {since})."

    # noinspection PyUnusedLocal
    @botcmd(admin_only=True)
    def restart(self, msg, args):
        """ Restart the bot. """
        self.send(msg.frm, "Deactivating all the plugins...")
        self._bot.plugin_manager.deactivate_all_plugins()
        self.send(msg.frm, "Restarting")
        self._bot.shutdown()
        global_restart()
        return "I'm restarting..."

    # noinspection PyUnusedLocal
    @arg_botcmd('--confirm', dest="confirmed", action="store_true",
                help="confirm you want to shut down", admin_only=True)
    @arg_botcmd('--kill', dest="kill", action="store_true",
                help="kill the bot instantly, don't shut down gracefully", admin_only=True)
    def shutdown(self, msg, confirmed, kill):
        """
        Shutdown the bot.

        Useful when the things are going crazy and you don't have access to the machine.
        """
        if not confirmed:
            yield "Please provide `--confirm` to confirm you really want me to shut down."
            return

        if kill:
            yield "Killing myself right now!"
            os.kill(os.getpid(), signal.SIGKILL)
        else:
            yield "Roger that. I am shutting down."
            os.kill(os.getpid(), signal.SIGINT)
